""" solar.py: Model for solar time series data"""

import numpy as np
import pandas as pd
from pmdarima import auto_arima
from statsmodels.tsa.statespace.sarimax import SARIMAX
from statsmodels.tsa.seasonal import STL, DecomposeResult

from .timeseries_model import TimeSeriesModel

import logging

logger = logging.getLogger(__name__)


def process_solar_series(solar_ts: pd.Series, suntimes: pd.DataFrame) -> pd.Series:
    # Remove negative values
    output_df = solar_ts.copy()

    # Convert sunrise and sunset to datetime
    suntimes["sunrise"] = pd.to_datetime(suntimes["sunrise"])
    suntimes["sunset"] = pd.to_datetime(suntimes["sunset"])

    output_df = output_df.to_frame().join(suntimes)
    output_df.loc[
        (output_df.index.hour < output_df["sunrise"].dt.hour)
        | (output_df.index.hour > output_df["sunset"].dt.hour),
        "value",
    ] = 0

    # Any negative value is zero
    output_df["value"] = np.maximum(output_df["value"], 0)
    # Drop sunrise and sunset columns
    output_df = output_df.drop(columns=["sunrise", "sunset"])

    # All negative values are set to zero
    output_df["value"] = np.maximum(output_df["value"], 0)
    return output_df


class SolarTSModel(TimeSeriesModel):
    def __init__(
        self,
    ) -> None:
        super().__init__()
        self._monthly_models: dict[int, SARIMAX] = {}
        self._predictions: pd.Series = pd.Series()
        self._pred_residuals: pd.Series = pd.Series()

        # Specific model parameters
        self.stl_seasonal_value = 24 * 7 - 1  # Must be an odd number
        self.stl_period_value = 24
        self.monthly_stl_results: dict[int, DecomposeResult] = {}

    @property
    def monthly_models(self) -> dict:
        if not self._is_fitted:
            raise ValueError("Model must be fitted first!")
        return self._monthly_models

    @property
    def predictions(self) -> pd.Series:
        if not self._is_fitted:
            raise ValueError("Model must be fitted first!")
        return self._predictions

    @property
    def pred_residuals(self) -> pd.Series:
        if not self._is_fitted:
            raise ValueError("Model must be fitted first!")
        return self._pred_residuals

    def _fit(
        self,
        target_column: str,
        arima_order: tuple[int, int, int],
        seasonal_order: tuple[int, int, int, int] = None,
        exog_vars: list[str] = None,
    ) -> None:
        # Check that self.data has sunrise and sunset columns
        if "sunrise" not in self.data.columns or "sunset" not in self.data.columns:
            raise ValueError("Data should have columns 'sunrise' and 'sunset'")

        self._pred_residuals = pd.Series()
        for month in self.months:
            logger.info(f"Fitting SARIMAX model for month {month}")
            monthly_y = self.data.loc[self.data.index.month == month, target_column]
            # We are interested in fitting a time series model to the
            # trend + residuals of the STL decomposition
            # The choice of seasonal argument is arbitrary to weekly patterns
            stl_model = STL(
                monthly_y,
                seasonal=self.stl_seasonal_value,
                period=self.stl_period_value,
            )
            stl_result = stl_model.fit()
            monthly_yt = monthly_y - stl_result.seasonal

            # SARIMAX
            exog_data = (
                self.data.loc[self.data.index.month == month, exog_vars]
                if exog_vars
                else None
            )
            sarimax_model = SARIMAX(
                monthly_yt,
                exog=exog_data,
                order=arima_order,
                seasonal_order=seasonal_order,
            ).fit(disp=True)

            # Store the models, and residuals
            self._monthly_models[month] = sarimax_model
            self.monthly_stl_results[month] = stl_result

            if self._pred_residuals.empty:
                self._pred_residuals = sarimax_model.resid
            else:
                self._pred_residuals = pd.concat(
                    [self._pred_residuals, sarimax_model.resid]
                )
        self._pred_residuals.name = "value"

    def _predict(self) -> pd.Series:
        self._predictions = pd.Series()
        for month in self.months:
            sarimax_model = self._monthly_models[month]
            # The SARIMAX model predicts yt, which is the trend and residuals of LOESS
            monthly_yt_pred = sarimax_model.predict()
            monthly_y_pred = monthly_yt_pred + self.monthly_stl_results[month].seasonal
            monthly_y_pred.name = "value"
            if self._predictions.empty:
                self._predictions = monthly_y_pred
            else:
                self._predictions = pd.concat([self._predictions, monthly_y_pred])

        # Post-processing involves removing negative values and
        # setting irradiance to zero during night hours
        suntimes = self.data.loc[:, ["sunrise", "sunset"]]
        self._predictions = process_solar_series(self._predictions, suntimes)

        return self._predictions

    def _get_synthetic(self, exog_data: pd.DataFrame, seed: int) -> pd.Series:
        synthetic_y = pd.Series()
        for month in self.months:
            sarimax_model = self._monthly_models[month]
            stl_result = self.monthly_stl_results[month]

            # Find the maximum number of days in the month
            max_day = self.data.loc[self.data.index.month == month].shape[0] // 24

            for day in range(1, max_day + 1):

                # Extract the start and end time for the day for indexing
                # the synthetic data
                start_time = self.data.loc[
                    (self.data.index.month == month) & (self.data.index.day == day)
                ].index[0]
                end_time = start_time + pd.Timedelta("23h")

                # This is the trend + residuals of the STL decomposition
                daily_exog_data = None
                if exog_data is not None:
                    daily_exog_data = exog_data.loc[start_time:end_time, :]

                daily_yt_syn = sarimax_model.simulate(
                    exog=daily_exog_data,
                    nsimulations=24,
                    anchor="end",
                    random_state=seed,
                )

                # Add the seasonal component from the STL decomposition to
                # get the synthetic data (irradiance)
                daily_y_syn = (
                    daily_yt_syn.values
                    + stl_result.seasonal.loc[start_time:end_time].values
                )
                daily_y_syn = pd.Series(
                    daily_y_syn,
                    name="value",
                    index=pd.date_range(start_time, end_time, freq="h"),
                )
                if synthetic_y.empty:
                    synthetic_y = daily_y_syn
                else:
                    synthetic_y = pd.concat([synthetic_y, daily_y_syn])

        # Post-processing involves removing negative values and
        # setting irradiance to zero during night hours
        suntimes = self.data.loc[:, ["sunrise", "sunset"]]
        synthetic_y = process_solar_series(synthetic_y, suntimes)
        return synthetic_y

    def _find_best_model(
        self,
        target_column: str,
        exog_vars: list[str],
        month_to_use: int,
        seed: int,
        suppress_warnings: bool,
    ) -> tuple[tuple[int, int, int], tuple[int, int, int, int]]:
        monthly_y = self.data.loc[self.data.index.month == month_to_use, target_column]
        monthly_exog = None
        if exog_vars:
            monthly_exog = self.data.loc[
                self.data.index.month == month_to_use, exog_vars
            ]

        # Find monthly_yt
        stl_model = STL(
            monthly_y,
            seasonal=self.stl_seasonal_value,
            period=self.stl_period_value,
        )
        stl_result = stl_model.fit()
        monthly_yt = monthly_y - stl_result.seasonal

        # Find the best model
        best_model = auto_arima(
            monthly_yt,
            X=monthly_exog,
            start_p=0,
            start_q=0,
            max_p=2,
            max_d=2,
            max_q=2,
            seasonal=False,
            information_criterion="aic",
            stepwise=True,
            suppress_warnings=suppress_warnings,
            error_action="warn",
            random_state=seed,
        )
        return best_model.order, best_model.seasonal_order



""" rounding_algo.py: Functions to perform iterative rounding.
"""

import gurobipy as gp
import numpy as np
import pandas as pd


def get_variables(model: gp.Model, target_varnames: list[str] = None) -> dict:
    """Extract non-binary variables from a Gurobi model.

    Args:
        model (gp.Model): The Gurobi model to extract variables from.
        target_varnames (list[str], optional):
            A list of variable name prefixes to include.
            If None, defaults to ["status"].

    Returns:
        dict: A dictionary mapping variable names to their corresponding
              non-binary values (v.X).

    """
    if target_varnames is None:
        target_varnames = ["status"]
    filtered_vars = {}
    for v in model.getVars():
        # Extract the prefix of the variable name (e.g., 'status' from 'status[1,2]')
        if v.varname.split("[")[0] in target_varnames:
            filtered_vars[v.varname] = v
    return filtered_vars


def find_fraction_vars(
    binary_vars: dict,
    atol: float = 1e-5,
) -> dict:
    """Return a list of variable names when their values are fractional."""
    fractional_vars = {}
    for varname in binary_vars:
        x_value = binary_vars[varname].X
        if not (np.isclose(x_value, 0, atol=atol) or np.isclose(x_value, 1, atol=atol)):
            fractional_vars[varname] = binary_vars[varname]
    return fractional_vars


def round_up(variable: gp.Var) -> None:
    variable.lb = 1
    variable.ub = 1


def round_down(variable: gp.Var) -> None:
    variable.lb = 0
    variable.ub = 0


def slow_rounding(
    fraction_vars: dict,
    threshold: float = 0,
) -> None:
    """Iteratively rounding variables with the largest value at each iteration.
    Values above the threshold are rounded up. Values below the threshold are rounded down.
    """
    max_value = max([v.X for v in fraction_vars.values()])
    for var in fraction_vars.values():
        if var.X == max_value:
            if max_value >= threshold:
                round_up(var)
            else:
                round_down(var)


def fast_rounding(fraction_vars: dict, threshold: float = 0) -> None:
    for bin_var in fraction_vars.values():
        if bin_var.X >= threshold:
            round_up(bin_var)
        else:
            round_down(bin_var)


def check_binary_values(var_dict: dict) -> bool:
    """
    Check if all variables in a dictionary have binary values (0 or 1).

    Args:
        var_dict (dict): A dictionary where keys are variable names and
                          values are gurobipy.Var objects.

    Returns:
        bool: True if all variables have binary values, False otherwise.
    """
    for var_name, var in var_dict.items():
        var_value = var.X
        if not (var_value == 0 or var_value == 1):
            print(f"Variable {var_name} has non-binary value: {var_value}")
            return False
    return True


def optimize_with_rounding(
    model: gp.Model,
    rounding_strategy: str,
    threshold: float,
    max_rounding_iter: int,
    mipgap: float,
    timelimit: int,
    num_threads: int,
    log_to_console: bool,
) -> tuple[gp.Model, float, int]:
    """
    Optimize a Gurobi model using iterative rounding with a given threshold.

    This function first relaxes the input model and then iteratively rounds
    fractional variables until an integer solution is found or the maximum
    number of iterations is reached.

    Args:
        model (gp.Model): The Gurobi model to optimize.
        threshold (float): The threshold for rounding fractional variables.
        max_rounding_iter (int): The maximum number of rounding iterations.
        log_to_console (bool): Whether to log optimization output to the console.
        mipgap (float): The relative MIP optimality gap.
        timelimit (int): The time limit for the optimization in seconds.
        num_threads (int): The number of threads to use for optimization.

    Returns:
        gp.Model: The optimized Gurobi model.
    """

    # First specify the model parameters
    model.Params.LogToConsole = log_to_console
    model.Params.MIPGap = mipgap
    model.Params.TimeLimit = timelimit
    model.Params.Threads = num_threads

    rounding_model = model.relax()
    rounding_model.Params.LogToConsole = False
    binary_vars = get_variables(rounding_model)

    rounding_optimization_time = 0
    for current_iter in range(max_rounding_iter):
        rounding_model.optimize()

        # Keep track of the optimization time
        rounding_optimization_time += rounding_model.runtime

        # Fixing variables can cause infeasibility
        if rounding_model.status == 3:
            print("\nPowNet: Rounding is infeasible. Use the MIP method.")
            break
        # The model should be feasible, but raise an error if not.
        elif rounding_model.status != 2:
            raise ValueError(f"Unrecognized model status: {rounding_model.status}")

        # Round variables and update the model
        fraction_vars = find_fraction_vars(binary_vars)

        # An empty dict means we have an integer solution.
        if len(fraction_vars) == 0:
            return rounding_model, rounding_optimization_time, current_iter

        if rounding_strategy == "slow":
            slow_rounding(fraction_vars=fraction_vars, threshold=threshold)
        else:
            fast_rounding(fraction_vars=fraction_vars, threshold=threshold)

        # Remove the rounded variables
        rounding_model.update()

    # If no integer solution is found after max_rounding_iter
    print(
        "\nPowNet: The rounding heuristic has terminated before finding an integer solution."
    )
    model.optimize()
    return model, None, None



""" kirsch_nowak.py
"""

import numpy as np
from scipy.linalg import cholesky, eig


def KNN_identification(Z, Qtotals, month, k=None):
    """
    Identifies K-nearest neighbors of Z in the historical annual data
    and computes the associated weights W.

    Args:
        Z: Synthetic datum (scalar)
        Qtotals: Total monthly flows at all sites for all historical months
                 within +/- 7 days of the month being disaggregated
        month: Month being disaggregated
        k: Number of nearest neighbors (by default k=n_year^0.5
           according to Lall and Sharma (1996))


    Returns:
        KNN_id: Indices of the first K-nearest neighbors of Z in the
                historical annual data z
        W: Nearest neighbors weights, according to Lall and Sharma (1996):
           W(i) = (1/i) / (sum(1/i))
    """

    # Ntotals is the number of historical monthly patterns used for disaggregation.

    Ntotals = Qtotals[month].shape[0]

    if k is None:
        K = round(np.sqrt(Ntotals))
    else:
        K = k

    # Nearest neighbors identification
    Nsites = Qtotals[month].shape[1]
    delta = np.zeros(Ntotals)

    for i in range(Ntotals):
        for j in range(Nsites):
            delta[i] += (Qtotals[month][i, j] - Z[0, 0, j]) ** 2

    Y = np.column_stack((np.arange(1, Ntotals + 1), delta))
    Y_ord = Y[Y[:, 1].argsort()]  # Sort by the second column (delta)
    KNN_id = Y_ord[:K, 0].astype(int)  # Extract the first column (indices)

    # Computation of the weights
    f = np.arange(1, K + 1)
    f1 = 1 / f
    W = f1 / np.sum(f1)

    return KNN_id, W


def chol_corr(Z):
    """
    Computes the Cholesky decomposition of the correlation matrix of the columns of Z.
    Attempts to repair non-positive-definite matrices.
    """

    R = np.corrcoef(Z, rowvar=False)  # Calculate the correlation matrix
    U, p = cholesky(R, lower=False)  # Attempt Cholesky decomposition

    while p > 0:  # If not positive definite, modify slightly
        k = min(
            [np.min(np.real(eig(R)[0])) - 1e-15, -1e-15]
        )  # Smallest eigenvalue or a small negative value
        R = R - k * np.eye(R.shape[0])
        R = R / R[0, 0]  # Rescale to get unit diagonal entries
        U, p = cholesky(R, lower=False)  # Retry Cholesky decomposition

    return U



"""model_formulation.py: Contains functions for constructing the objective function and constraints of the optimization model."""

from pownet.core.input import SystemInput

""" Functions for the objective function
"""


def get_thermal_fixed_coeff(
    timesteps: range,
    thermal_units: list,
    thermal_fixed_cost: dict,
    thermal_rated_capacity: dict,
) -> dict:
    """Fixed cost is a function of rated capacity and fixed cost per unit"""
    return {
        (unit, t): thermal_rated_capacity[unit] * thermal_fixed_cost[unit]
        for t in timesteps
        for unit in thermal_units
    }


def get_thermal_opex_coeff(
    step_k: int,
    timesteps: range,
    thermal_units: list,
    thermal_opex: dict,
    fuel_contracts: dict,
    contract_costs: dict,
    thermal_heat_rate: dict,
) -> dict:
    """Variable cost is a function of fuel cost, heat rate, and opex"""
    # TODO: thermal_fuel_cost should be a time-dependent parameter
    return {
        (unit, t): (
            contract_costs[(fuel_contracts[unit], t + (step_k - 1) * 24)]
            * thermal_heat_rate[unit]
        )
        + thermal_opex[unit]
        for t in timesteps
        for unit in thermal_units
    }


def get_thermal_startup_coeff(
    timesteps: range,
    thermal_units: list,
    thermal_startup_cost: dict,
    thermal_rated_capacity: dict,
) -> dict:
    """Startup cost is a function of rated capacity and startup cost per unit"""
    return {
        (unit, t): thermal_rated_capacity[unit] * thermal_startup_cost[unit]
        for t in timesteps
        for unit in thermal_units
    }


def get_marginal_cost_coeff(
    step_k: int,
    timesteps: range,
    units: list,
    nondispatch_contracts: dict,
    contract_costs: dict,
) -> dict:
    """
    Generic helper function to calculate coefficients based on marginal cost or a similar attribute for a list of units

    Args:
        units: The list of units to calculate coefficients for
        attribute: The attribute name in inputs to use for cost calculation

    Returns
        A dictionary mapping (unit, t) tuples to the calculated coefficients
    """
    return {
        (unit, t): contract_costs[(nondispatch_contracts[unit], t + (step_k - 1) * 24)]
        for t in timesteps
        for unit in units
    }



""" This script provides an example of how to run PowNet 2.0.
"""

import os
from pownet.core import Simulator


def main():

    # --------- User inputs

    input_folder = "..//model_library"
    output_folder = "..//temptemp"

    model_name = "dummy"
    model_year = 2016

    # Simulation parameters
    sim_horizon = 24
    steps_to_run = 2
    solver = "gurobi"  # or highs

    # --------- End of user inputs

    # Run the simulation
    simulator = Simulator(
        input_folder=input_folder,
        model_name=model_name,
        model_year=model_year,
    )

    simulator.run(
        sim_horizon=sim_horizon,
        steps_to_run=steps_to_run,
        solver=solver,
    )

    # Write the simulation results
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    simulator.write_results(output_folder)

    # Plot the results
    simulator.plot_fuelmix("bar", output_folder)
    simulator.plot_thermal_units(output_folder)


if __name__ == "__main__":
    main()



""" timeseries_utils.py: Utility functions for timeseries data """

import numpy as np
import pandas as pd
from sklearn.preprocessing import PowerTransformer


def transform_log1p(y: pd.Series, B1_y: pd.Series):
    # Take difference
    y = y - B1_y
    # Apply the transformation
    y = np.log1p(y)
    return y


def inverse_log1p(y_pred: pd.Series, B1_y: pd.Series) -> pd.Series:
    # invert the transformation
    y_pred = np.expm1(y_pred)
    # Convert negative values to 0
    y_pred = np.maximum(y_pred, 0)
    # Add the last observed value
    y_pred = y_pred + B1_y
    return y_pred


def transform_yeojohnson(
    y: pd.Series, B1_y: pd.Series
) -> tuple[pd.Series, PowerTransformer]:
    """Differencing the data and apply the Yeo-Johnson transformation to the data"""
    # Take difference and discard the first observation
    transformed_y = (y - B1_y).dropna()
    # Apply the transformation
    power_transformer = PowerTransformer(method="yeo-johnson")
    power_transformer.fit(transformed_y.values.reshape(-1, 1))
    transformed_y = power_transformer.transform(transformed_y.values.reshape(-1, 1))
    return transformed_y, power_transformer


def inverse_yeojohnson(
    y_pred: np.array,
    B1_y: pd.Series,
    initial_value: float,
    power_transformer: PowerTransformer,
) -> pd.Series:
    """Invert the Yeo-Johnson transformation"""
    # invert the transformation
    y_pred = power_transformer.inverse_transform(y_pred.reshape(-1, 1))
    y_pred = pd.Series(y_pred.flatten(), index=B1_y.index)
    # Add value from the previous timestep because the data was differenced
    y_pred = y_pred + B1_y
    # Insert the first observed value to the beginning of the series
    # because the first value was removed during differencing
    temp_series = pd.Series([initial_value])
    temp_series.index = [B1_y.index[0] - 1]
    y_pred = pd.concat([temp_series, y_pred])
    return y_pred


def post_process_solar(
    solar_data: pd.DataFrame, value_cols: list[str], sunrise_time: str, sunset_time: str
):
    # Ensure values are zero outside sunset and sunrise times
    solar_data.loc[
        (solar_data["datetime"].dt.time < pd.to_datetime(sunrise_time).time())
        | (solar_data["datetime"].dt.time > pd.to_datetime(sunset_time).time()),
        value_cols,
    ] = 0

    # Ensure no negative values, which can happen with ARIMA predictions
    solar_data[value_cols] = solar_data[value_cols].clip(lower=0)



""" This is test_model.py
"""

import os
import unittest
from unittest.mock import patch, MagicMock, call

import gurobipy as gp
import pandas as pd
from pownet.modeling import PowerSystemModel


class TestPowerSystemModel(unittest.TestCase):
    def setUp(self):
        """Create a simple optimization problem of two variables and a constraint
        min -x - 2y
        s.t. x + y >= 1
        x, y >= 0

        solution: x = 0, y = 1, objval = -2
        """
        self.gurobi_instance = gp.Model()
        x = self.gurobi_instance.addVar(name="x", lb=0)
        y = self.gurobi_instance.addVar(name="y", lb=0)
        self.gurobi_instance.setObjective(-x - 2 * y, gp.GRB.MINIMIZE)
        self.gurobi_instance.addConstr(x + y <= 1)
        self.gurobi_instance.update()
        self.psm = PowerSystemModel(self.gurobi_instance)

        self.infeasible_gurobi_instance = gp.Model()
        x = self.infeasible_gurobi_instance.addVar(name="x", lb=0)
        y = self.infeasible_gurobi_instance.addVar(name="y", lb=0)
        self.infeasible_gurobi_instance.setObjective(x + y, gp.GRB.MINIMIZE)
        self.infeasible_gurobi_instance.addConstr(x + y <= -1)
        self.infeasible_gurobi_instance.update()
        self.infeasible_psm = PowerSystemModel(self.infeasible_gurobi_instance)

        self.expected_solution = pd.DataFrame(
            {"varname": ["x", "y"], "value": [0.0, 1.0]}
        )

    def test_optimize_gurobi(self):
        self.psm.optimize(solver="gurobi", log_to_console=False)
        self.psm.check_feasible()
        self.assertEqual(self.psm.get_objval(), -2)

    def test_optimize_highs(self):
        self.psm.optimize(solver="highs", log_to_console=False)
        info = self.psm.model.getInfo()
        self.psm.check_feasible()
        self.assertEqual(info.objective_function_value, -2)

    def test_optimize_invalid_solver(self):
        with self.assertRaises(ValueError):
            self.psm.optimize(solver="invalid_solver")

    def test_check_infeasible_gurobi(self):
        self.infeasible_psm.optimize(solver="gurobi", log_to_console=False)
        self.assertFalse(self.infeasible_psm.check_feasible())

    def test_get_solution_gurobi(self):

        self.psm.optimize(solver="gurobi", log_to_console=False)
        pd.testing.assert_frame_equal(
            pd.DataFrame(self.psm.get_solution()), self.expected_solution
        )

    def test_get_solution_highs(self):
        self.psm.optimize(solver="highs", log_to_console=False)
        pd.testing.assert_frame_equal(
            pd.DataFrame(self.psm.get_solution()), self.expected_solution
        )


if __name__ == "__main__":
    unittest.main()



""" reservoir_operator.py: Defines the ReservoirOperator class """

import os

import numpy as np
import pandas as pd

from pownet.folder_utils import get_reservoir_file

from .basin import Basin
from .reservoir import Reservoir
from .reservoir_functions import convert_to_hourly_hydropower


class ReservoirOperator:
    def __init__(
        self,
        model_name: str,
        sim_horizon: int = 365,
    ) -> None:
        """A class that manages the operation of multiple basins."""

        self.model_name: str = model_name
        self.sim_horizon: int = sim_horizon

        # Basin information
        self.basin_names: list[str] = []
        self.reservoir_names: list[str] = []

        # Basins in the system
        self.basins: list[Basin] = []
        self.reservoirs: list[Reservoir] = []

        # Hydropower timeseries
        self.daily_hydropower: pd.DataFrame = pd.DataFrame()
        self.hourly_hydropower: pd.DataFrame = pd.DataFrame()

    def load_csv(self) -> None:
        """
        Load information from a CSV file.
        """
        # Basin names
        self.basin_names = (
            pd.read_csv(
                get_reservoir_file(self.model_name, "reservoir.csv"),
                header=0,
            )["basin"]
            .unique()
            .tolist()
        )
        # All reservoir names in the system
        self.reservoir_names.extend(
            name for basin in self.basins for name in basin.reservoir_names
        )

    def simulate(self):
        """Simulate the operation of the reservoirs."""
        # Create Basin objects
        for basin_name in self.basin_names:
            basin = Basin(model_name=self.model_name, basin_name=basin_name)
            basin.load_csv()
            basin.simulate()
            self.basins.append(basin)

            # Get the reservoirs in the basin
            self.reservoirs.extend(
                [res for res_lists in basin.reservoirs.values() for res in res_lists]
            )

            # Get the hydropower timeseries
            self.daily_hydropower = pd.concat(
                [self.daily_hydropower, basin.get_basin_hydropower(timestep="daily")],
                axis=1,
            )
            self.hourly_hydropower = convert_to_hourly_hydropower(self.daily_hydropower)

    def get_daily_hydropower(self) -> pd.DataFrame:
        return self.daily_hydropower

    def get_hourly_hydropower(self) -> pd.DataFrame:
        return self.hourly_hydropower

    def export_hydropower_csv(self, timestep: str, output_folder: str) -> None:
        csv_name = "pownet_hydropower.csv"

        # Check if the timestep is valid
        if timestep not in ["daily", "hourly"]:
            raise ValueError("Unknown timestep")

        # Map timestep to the corresponding DataFrame
        hydropower_data = {
            "daily": self.daily_hydropower,
            "hourly": self.hourly_hydropower,
        }
        # Export the DataFrame to CSV
        hydropower_data[timestep].to_csv(os.path.join(output_folder, csv_name))

    def get_plots(self, output_folder: str = None) -> None:
        """Plot daily storage, release, and spill of each reservoir."""
        for reservoir in self.reservoirs:
            reservoir.plot_state(output_folder=output_folder)

    def export_state(self, output_folder: str = None) -> None:
        """Export the daily values of storage, release, and spill of each reservoir."""
        for reservoir in self.reservoirs:
            df = pd.DataFrame(
                {
                    "inflow": reservoir.inflow,
                    "upstream_flow": reservoir.upstream_flow,
                    "storage": reservoir.storage,
                    "release": reservoir.release,
                    "spill": reservoir.spill,
                    "level": reservoir.level,
                    "mid_level": reservoir.mid_level,
                    "min_flow": reservoir.min_flow,
                    "target_storage": reservoir.target_storage,
                }
            )
            df.to_csv(
                os.path.join(output_folder, f"{reservoir.name}_state.csv"),
                index=False,
            )

    def export_reop_state(self, output_folder: str = None) -> None:
        for reservoir in self.reservoirs:
            df = pd.DataFrame(
                {
                    "inflow": reservoir.inflow,
                    "upstream_flow": reservoir.reop_upstream,
                    "storage": reservoir.reop_storage,
                    "release": reservoir.reop_release,
                    "spill": reservoir.reop_spill,
                    "level": reservoir.reop_level,
                    "mid_level": reservoir.mid_level,
                    "min_flow": reservoir.min_flow,
                    "target_storage": reservoir.target_storage,
                }
            )
            df.to_csv(
                os.path.join(output_folder, f"reop_{reservoir.name}_state.csv"),
                index=False,
            )

    def reoperate_basins(
        self, day: int, daily_dispatch: dict[str, float]
    ) -> dict[str, float]:
        """Reoperate the reservoirs based on the daily dispatch of the power system model.
        Note that we don't reoperate on the first day of the simulation period.
        """
        daily_hydropower_capacity = {}

        for basin in self.basins:
            # Reset reoperated downstream flow
            basin.init_reop_cascade_flow()
            for cascade_lv in range(basin.max_cascade + 1):
                for reservoir in basin.reservoirs[cascade_lv]:
                    reservoir_daily_dispatch = daily_dispatch[reservoir.name]
                    daily_hydropower_capacity[reservoir.name] = reservoir.reoperate(
                        day=day,
                        daily_dispatch=reservoir_daily_dispatch,
                        upstream_flow=basin.cascade_flow.loc[day, cascade_lv],
                    )
                    # Update downstream flow
                    basin.cascade_flow.loc[day, cascade_lv + 1] += (
                        +reservoir.reop_release[day] + reservoir.reop_spill[day]
                    )

        return daily_hydropower_capacity



""" Unit tests for the ModelBuilder class.
"""

import os
import unittest
from pownet.core import ModelBuilder, SystemInput
from pownet.folder_utils import get_model_dir


class TestModelBuilder(unittest.TestCase):
    """Unless otherwise stated, use "dummy_trade" over 24-hr as the test case."""

    def setUp(self) -> None:
        # Load the test data
        test_model_library_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "test_model_library")
        )

        self.inputs = SystemInput(
            input_folder=test_model_library_path,
            model_name="dummy",
            year=2016,
            sim_horizon=24,
        )
        self.inputs.load_data()
        self.inputs.check_data()
        # Create a model builder object
        self.model_builder = ModelBuilder(self.inputs)

    def test_init(self):
        self.assertEqual(
            self.model_builder.timesteps, range(1, self.inputs.sim_horizon + 1)
        )


if __name__ == "__main__":
    unittest.main()



from pownet.stochastic.solar import SolarTSModel
import pandas as pd
import matplotlib.pyplot as plt

import logging

# Show info
logging.basicConfig(level=logging.INFO)

# %% Solar
data = pd.read_csv("../temp/merra_2019.csv")

solar_model = SolarTSModel()
solar_model.load_data(data)

solar_model.fit(target_column="ground_irradiance", arima_order=(2, 1, 2))
predictions = solar_model.predict()

resids = solar_model.pred_residuals

synthetic = solar_model.get_synthetic()

duration = 30
# plt.plot(predictions[: 24 * 3], label="Predictions")
plt.plot(synthetic[: 24 * duration], label="Synthetic")
plt.plot(
    data.set_index("datetime")["ground_irradiance"].iloc[: 24 * duration],
    label="Actual",
)
plt.legend()
plt.show()



import unittest
import pandas as pd

from pownet.data_utils import calc_remaining_on_duration, calc_remaining_off_duration


class TestCalcMinOnlineDuration1(unittest.TestCase):

    def test_startup_within_horizon(self):
        """Test case where a unit has a startup within the simulation horizon.
        - Simulation horizon of 10 hours.
        - A minimum online duration (TU) of 3 hours
        - Latest start at hour 8.
        - The remaining minimum online duration is 0.
        """
        solution = pd.DataFrame(
            {
                "node": ["Unit1"] * 10 + ["Unit2"] * 10,
                "timestep": [x for x in range(1, 11)] + [x for x in range(1, 11)],
                "value": [0, 0, 0, 1, 1, 0, 0, 1, 0, 0]
                + [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                "vartype": ["startup"] * 10 + ["shutdown"] * 10,
            }
        )
        sim_horizon = 10
        thermal_units = ["Unit1"]
        TU = {"Unit1": 3}

        expected_output = {
            "Unit1": 0,
        }  # Unit1 started recently, Unit2 has some time left
        result = calc_remaining_on_duration(solution, sim_horizon, thermal_units, TU)
        self.assertEqual(result, expected_output)

    def test_startup_outside_horizon(self):
        """Test case where a unit has a startup within the simulation horizon.

        Unit1
        -------
        - A minimum online duration (TU) of 5 hours
        - Latest start at hour 8.
        - The remaining minimum online duration is 2.

        Unit2
        -------
        - A minimum online duration (TU) of 2 hours
        - Latest start at hour 10.
        - The remaining minimum online duration is 1.


        """
        solution = pd.DataFrame(
            {
                "node": ["Unit1"] * 10 + ["Unit2"] * 10,
                "timestep": [x for x in range(1, 11)] + [x for x in range(1, 11)],
                "value": [0, 0, 0, 1, 1, 0, 0, 1, 0, 0]
                + [0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                "vartype": ["startup"] * 20,
            }
        )
        sim_horizon = 10
        thermal_units = ["Unit1", "Unit2"]
        TU = {"Unit1": 5, "Unit2": 2}

        expected_output = {
            "Unit1": 2,
            "Unit2": 1,
        }  # Unit1 started recently, Unit2 has some time left
        result = calc_remaining_on_duration(solution, sim_horizon, thermal_units, TU)
        self.assertEqual(result, expected_output)


class TestCalcMinOnlineDuration2(unittest.TestCase):

    def test_shutdown_within_horizon(self):
        """Test case where a unit has a shutdown within the simulation horizon.
        - Simulation horizon of 10 hours.
        - A minimum offline duration (TD) of 1 hours
        - Latest shutdown at hour 10.
        - The remaining minimum offline duration is 0.
        """
        solution = pd.DataFrame(
            {
                "node": ["Unit1"] * 10 + ["Unit2"] * 10,
                "timestep": [x for x in range(1, 11)] + [x for x in range(1, 11)],
                "value": [0, 0, 0, 1, 1, 0, 0, 1, 1, 1]
                + [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                "vartype": ["dummy_var"] * 10 + ["shutdown"] * 10,
            }
        )
        sim_horizon = 10
        thermal_units = ["Unit2"]  # Only test for Unit2
        TD = {"Unit2": 1}

        expected_output = {
            "Unit2": 0,
        }  # Unit1 started recently, Unit2 has some time left
        result = calc_remaining_off_duration(solution, sim_horizon, thermal_units, TD)
        self.assertEqual(result, expected_output)

    def test_shutdown_outside_horizon(self):
        """Test case where a unit has a shutdown within the simulation horizon.

        Unit1
        -------
        - A minimum offline duration (TD) of 5 hours
        - Latest shutdown at hour 6.
        - The remaining minimum offline duration is 0.

        Unit2
        -------
        - A minimum offline duration (TD) of 2 hours
        - Latest shutdown at hour 10.
        - The remaining minimum offline duration is 1.


        """
        solution = pd.DataFrame(
            {
                "node": ["Unit1"] * 10 + ["Unit2"] * 10,
                "timestep": [x for x in range(1, 11)] + [x for x in range(1, 11)],
                "value": [0, 0, 0, 0, 0, 1, 0, 0, 0, 0]
                + [0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                "vartype": ["shutdown"] * 20,
            }
        )
        sim_horizon = 10
        thermal_units = ["Unit1", "Unit2"]
        TD = {"Unit1": 5, "Unit2": 2}

        expected_output = {
            "Unit1": 0,
            "Unit2": 1,
        }
        result = calc_remaining_off_duration(solution, sim_horizon, thermal_units, TD)
        self.assertEqual(result, expected_output)


if __name__ == "__main__":
    unittest.main()



""" timeseries_model.py: Abstract class for time series models."""

from abc import ABC, abstractmethod

import pandas as pd


class TimeSeriesModel(ABC):
    def __init__(
        self,
    ) -> None:
        self._is_fitted: bool = False
        self._is_loaded: bool = False

        # DataFrame to be loaded
        self.data: pd.DataFrame = pd.DataFrame()
        self.months: list = []  # Months are labeled from 1 to 12
        self.exog_vars: list[str] = None

    @property
    @abstractmethod
    def monthly_models(self) -> dict:
        pass

    @property
    @abstractmethod
    def predictions(self) -> pd.Series:
        pass

    @property
    @abstractmethod
    def pred_residuals(self) -> pd.Series:
        pass

    def load_data(self, data: pd.DataFrame) -> None:
        """Load the time series data"""
        # Check that there are columns called 'datetime' and 'value'
        if "datetime" not in data.columns:
            raise ValueError("Data should have columns 'datetime'")
        self.data = data
        # Convert 'datetime' column to datetime type
        self.data["datetime"] = pd.to_datetime(self.data["datetime"])
        self.months = self.data["datetime"].dt.month.unique().tolist()
        # Downstream code expects months to be sorted
        self.months.sort()
        # Time series modeling requires datetime to be the index
        self.data = self.data.set_index("datetime")
        self.data.index.freq = "h"  # Requires hourly time series
        self._is_loaded = True

    def fit(
        self,
        target_column: str,
        arima_order: tuple[int, int, int],
        seasonal_order: tuple[int, int, int, int] = None,
        exog_vars: list[str] = None,
    ) -> None:
        """Fit the model to the data

        Args:
            target_column (str): Target column to fit the model
            arima_order (tuple[int, int, int]): ARIMA order
            seasonal_order (tuple[int, int, int, int]): Seasonal order
            exog_vars (list[str], optional): List of exogenous variables. Defaults to None.
        """
        self._fit(
            target_column=target_column,
            arima_order=arima_order,
            seasonal_order=seasonal_order,
            exog_vars=exog_vars,
        )
        self._is_fitted = True
        self.exog_vars = exog_vars

    def predict(self) -> pd.Series:
        """Return the predictions"""
        if not self._is_fitted:
            raise ValueError("Model must be fitted before making predictions.")
        return self._predict()

    def get_synthetic(
        self,
        exog_data: pd.DataFrame = None,
        seed: int = None,
    ) -> pd.Series:
        """
        Create synthetic time series.

        Args:
            n_samples (int, optional): Number of samples to generate. Defaults to 1.
            exog_data (pd.DataFrame, optional): Exogenous variables. Defaults to None.
            seed (int, optional): Random seed. Defaults to None.

        Returns:
            pd.Series: Synthetic time series data.

        Raises:
            ValueError: If the model is not fitted.
            ValueError: If exogenous variables are not in the data.
            ValueError: If exogenous data index is not equal to the time series data index.
        """
        if not self._is_fitted:
            raise ValueError("Model must be fitted before creating synthetic data.")
        if exog_data is not None:
            if not all([var in exog_data.columns for var in self.exog_vars]):
                raise ValueError("Exogenous variables should be in the data.")
            if not exog_data.index.equals(self.data.index):
                raise ValueError(
                    "Exogenous data should have the same index as the time series data."
                )

        return self._get_synthetic(exog_data=exog_data, seed=seed)

    def find_best_model(
        self,
        target_column: str,
        exog_vars: list[str] = None,
        month_to_use: int = 1,
        seed: int = None,
        suppress_warnings: bool = False,
    ) -> tuple[tuple[int, int, int], tuple[int, int, int, int]]:
        """Find the best model for the time series data

        Args:
            target_column (str): Target column to fit the model
            exog_vars (list[str], optional): List of exogenous variables. Defaults to None.
            month_to_use (int, optional): Month to use for finding the best model. Defaults to 1.
            seed (int, optional): Random seed. Defaults to 112.
            suppress_warnings (bool, optional): Suppress warnings. Defaults to False.

        Returns:
            tuple[tuple[int, int, int], tuple[int, int, int, int]]: Best model SARIMA parameters

        Raises:
            ValueError: If data is not loaded
        """
        if not self._is_loaded:
            raise ValueError("Data must be loaded first.")
        return self._find_best_model(
            target_column=target_column,
            exog_vars=exog_vars,
            month_to_use=month_to_use,
            seed=seed,
            suppress_warnings=suppress_warnings,
        )

    @abstractmethod
    def _fit(
        self,
        target_column: str,
        arima_order: tuple[int, int, int],
        seasonal_order: tuple[int, int, int, int],
        exog_vars: list[str],
    ) -> None:
        pass

    @abstractmethod
    def _predict(self) -> pd.Series:
        pass

    @abstractmethod
    def _get_synthetic(self, exog_vars: list[str], seed: int) -> pd.Series:
        pass

    @abstractmethod
    def _find_best_model(
        self,
        target_column: str,
        exog_vars: list[str],
        month_to_use: int,
        seed: int,
        suppress_warnings: bool,
    ) -> tuple[tuple[int, int, int], tuple[int, int, int, int]]:
        pass



""" basin.py: Basin class
"""

import os

import pandas as pd

from pownet.folder_utils import get_reservoir_file
from .reservoir import Reservoir


class Basin:
    def __init__(
        self,
        model_name: str,
        basin_name: str,
        sim_horizon: int = 365,
    ) -> None:
        """
        A class to manage reservoirs in a basin.
        """

        self.model_name: str = model_name
        self.basin_name: str = basin_name
        self.sim_horizon: int = sim_horizon

        # Reservoir names in the basin
        self.reservoir_names: list[str] = []
        # The largest cascade level in the basin
        self.max_cascade: int = 0
        # Map cascade level to its list of reservoir names {cascade_lvl: [name1, name2]}
        self.map_cascade: dict[int, list[str]] = {}

        # Upstream flow for each cascade level
        # Each column represents outflow from reservoirs at the previous cascade level
        # Ending index has +2 because
        # (+1) Python indexing excludes the last index
        # (+1) Outflow from the last cascade level
        self.cascade_flow = pd.DataFrame()
        self.reop_cascade_flow = pd.DataFrame()

        # Dict of reservoir objects based on their cascasde level
        # {cascade_level: [Reservoir1, Reservoir2]}
        self.reservoirs: dict[int, Reservoir] = {}

    def load_csv(self) -> None:
        """Load Basin information from a CSV file."""

        # User must specify reservoirs and their parameters in a CSV file
        reservoirs = pd.read_csv(get_reservoir_file(self.model_name, "reservoir.csv"))

        # Filter to only reservoirs of the basin
        if self.basin_name not in reservoirs.basin.unique():
            raise ValueError("Basin not found. Check basin name.")
        reservoirs = reservoirs[reservoirs["basin"] == self.basin_name]
        self.reservoir_names = reservoirs["name"].tolist()

        # Map cascade level to its list of reservoir names {cascade_lvl: [name1, name2]}
        # Cascade level starts from 0 (most upstream)
        self.map_cascade = (
            reservoirs[["cascade_level", "name"]]
            .groupby("cascade_level")["name"]
            .apply(list)
            .to_dict()
        )

        # Largest cascade level in the basin
        self.max_cascade = reservoirs["cascade_level"].max()
        self.reservoirs: dict[int, Reservoir] = {
            i: [] for i in range(self.max_cascade + 1)
        }
        self.cascade_flow = pd.DataFrame(
            0, columns=range(self.max_cascade + 2), index=range(1, self.sim_horizon + 1)
        )

    def simulate(self) -> None:
        """Simulate the basin to fill in values of reservoirs."""
        # Initialize the upstream flow for each cascade level
        for cascade_lv in range(self.max_cascade + 1):
            reservoir_names = self.map_cascade[cascade_lv]
            for reservoir_name in reservoir_names:
                reservoir = Reservoir(
                    model_name=self.model_name,
                    reservoir_name=reservoir_name,
                    sim_horizon=self.sim_horizon,
                    upstream_flow=self.cascade_flow.loc[:, cascade_lv],  # Upstream flow
                )
                reservoir.load_from_csv()
                reservoir.simulate()
                self.reservoirs[cascade_lv].append(reservoir)

                # The outflow from this reservoir is becomes part of inflow
                # of the next cascade level. Hence, add 1 to the cascade level.
                self.cascade_flow.loc[:, cascade_lv + 1] += (
                    +reservoir.release + reservoir.spill
                )

    def _get_hydropower_data(self, get_hydropower_func) -> pd.DataFrame:
        """Helper function to get hydropower data from reservoirs sharing
        the same cascade level."""
        df = pd.DataFrame()
        for cascade_level in range(self.max_cascade + 1):
            reservoirs = self.reservoirs[cascade_level]
            temp_df = pd.DataFrame(
                {
                    reservoir.name: get_hydropower_func(reservoir)
                    for reservoir in reservoirs
                }
            )
            df = pd.concat([df, temp_df], axis=1)
        return df

    def get_basin_hydropower(self, timestep: str) -> pd.DataFrame:
        """Return a dataframe of hydropower by each reservoir.

        Args:
            timestep (str): The timestep of the data. Either "daily" or "hourly".

        Returns:
            pd.DataFrame: A dataframe with hydropower data
        """
        # Map timestep to the corresponding function
        hydropower_funcs = {
            "daily": lambda reservoir: reservoir.get_daily_hydropower(),
            "hourly": lambda reservoir: reservoir.get_hourly_hydropower(),
        }

        # Check if the timestep is valid
        if timestep not in hydropower_funcs:
            raise ValueError("Unknown timestep")

        # Get the hydropower data using the mapped function
        return self._get_hydropower_data(hydropower_funcs[timestep])

    def init_reop_cascade_flow(self) -> None:
        """Initialize cascade flow for the reservoir operation."""
        self.reop_cascade_flow = pd.DataFrame(
            0, columns=range(self.max_cascade + 2), index=range(1, self.sim_horizon + 1)
        )



""" reservoir_functions.py: Functions for operating reservoirs."""

import numpy as np
import pandas as pd


def adjust_hydropeaking(
    release: float,
    release_t0: float,
    max_release: float,
    min_release: float,
    hydropeak_factor: float = 0.15,
) -> float:
    """
    Adjust water release by considering hydropeaking and minimum environmental flow.
    The change in release is limited to hydropeak_factor times the maximum release.
    Also, the release cannot be lower than the minimum release or higher than the maximum release.

    Args:
        release (float): The release for the current day. Can be both negative and positive
        release_t0 (float): The release for the previous day. Can be both negative and positive
        max_release (float): The maximum release as a positive value
        min_release (float): The minimum release as a positive value
        hydropeak_factor (float): The factor to limit the change in release. Default is 0.15.

    Returns:
        float: The adjusted release.

    """
    # Calculate the difference between the current and previous release.
    # The sign is handled later in the code.
    release_change = release - release_t0

    # The rate of change is limited to hydropeak_factor
    change_limit = hydropeak_factor * max_release

    # Case 1: The difference is positive
    if release_change > 0:
        release_change = min(change_limit, release_change)

    # Case 2: The difference is negative, limit the decrease
    else:
        release_change = max(-change_limit, release_change)

    # Adjust the release based on the hydropeaking factor
    adj_release = release_t0 + release_change

    # Release is also bounded by min_release and  max_release
    adj_release = max(min_release, adj_release)
    adj_release = min(max_release, adj_release)
    return adj_release


def calc_min_environ_flow(
    inflow: float,
    mean_annual_flow: float,
    max_release: float,
) -> float:
    """Tdetermine the minimum amount of water that should be released
    from a reservoir to maintain the health of the downstream ecosystem.

    Example of three cases:

    1) If the inflow is less than 40% of the mean annual flow,
    the minimum flow is set to 60% of the inflow.
    This ensures that a reasonable portion of the limited water
    is still released to support the ecosystem.

    2) If the inflow is greater than 80% of the mean annual flow,
    the minimum flow is 30% of the inflow. a smaller percentage is
    released since the ecosystem is likely receiving ample water already.

    3) Otherwise, the minimum flow is 45% of the inflow.

    Args:
        inflow (float): The inflow to the reservoir
        mean_annual_flow (float): The mean annual flow
        max_release (float): The maximum release

    Returns:
        float: The minimum environmental flow
    """
    lower_maf_fraction = 0.4
    upper_maf_fraction = 0.8

    low_flow_scenario = lower_maf_fraction * mean_annual_flow
    upper_maf_fraction = upper_maf_fraction * mean_annual_flow

    small_fraction = 0.3
    medium_fraction = 0.45
    large_fraction = 0.6

    # Also need to ensure that the minimum environmental flow is less than the maximum release
    if inflow <= low_flow_scenario:
        return min(large_fraction * inflow, max_release)
    elif inflow > upper_maf_fraction:
        return min(small_fraction * inflow, max_release)
    else:
        return min(medium_fraction * inflow, max_release)


def calc_min_flow(
    inflow: pd.Series, mean_annual_flow: pd.Series, max_release: float
) -> None:
    """Find the minimum environmental flow for each day."""
    df = pd.DataFrame({"inflow": inflow, "mean_annual_flow": mean_annual_flow})
    min_flow = df.apply(
        lambda x: calc_min_environ_flow(
            inflow=x.inflow,
            mean_annual_flow=x.mean_annual_flow,
            max_release=max_release,
        ),
        axis=1,
    )
    return min_flow


def calc_target_level(
    min_day: int,
    max_day: int,
    min_level: float,
    max_level: float,
) -> pd.Series:
    """Calculate the target level for each day based on linear interpolation.

    Args:
        min_day (int): The day when the target level is at its minimum
        max_day (int): The day when the target level is at its maximum
        min_level (float): The minimum target level
        max_level (float): The maximum target level

    Returns:
        pd.Series: The target level for each day of the year
    """

    days_in_year = 365

    def _interpolate_between_min_and_max(day: int) -> float:
        """Interpolate target level between min_level and max_level."""
        return ((day - min_day) / (max_day - min_day)) * (
            max_level - min_level
        ) + min_level

    def _interpolate_after_max(day: int) -> float:
        """Interpolate target level after max_day."""
        return ((days_in_year - day + min_day) / (days_in_year - max_day + min_day)) * (
            max_level - min_level
        ) + min_level

    def _interpolate_before_min(day: int) -> float:
        """Interpolate target level before min_day."""
        return ((min_day - day) / (days_in_year - max_day + min_day)) * (
            max_level - min_level
        ) + min_level

    target_level = pd.Series(
        [
            (
                _interpolate_between_min_and_max(day)
                if min_day <= day <= max_day
                else (
                    _interpolate_after_max(day)
                    if day > max_day
                    else _interpolate_before_min(day)
                )
            )
            for day in range(1, days_in_year + 1)
        ],
        index=range(1, days_in_year + 1),
        dtype=float,
    )
    return target_level


def calc_target_storage(
    target_level: pd.Series,
    min_level: float,
    max_level: float,
    max_storage: float,
) -> pd.Series:
    """Calculate the target storage for each day based on linear interpolation.

    Args:
        target_level (pd.Series): The target level for each day
        min_level (float): The minimum level
        max_level (float): The maximum level
        max_storage (float): The maximum storage

    Returns:
        pd.Series: The target storage for each day
    """
    return ((target_level - min_level) / (max_level - min_level)) * max_storage


def calc_level_from_storage(
    storage: pd.Series,
    min_level: float,
    max_level: float,
    max_storage: float,
) -> pd.Series:
    """Calculate the level for each day based on linear interpolation.

    Args:
        storage (pd.Series): The storage for each day
        min_level (float): The minimum level
        max_level (float): The maximum level
        max_storage (float): The maximum storage

    Returns:
        pd.Series: The level for each day
    """
    return ((storage / max_storage) * (max_level - min_level)) + min_level


def calc_hourly_hydropower(
    release: pd.Series,
    mid_level: pd.Series,
    max_generation: float,
    turbine_factor: float,
    max_head: float,
    max_level: float,
) -> pd.Series:
    """Calculate hourly hydropower generation from release and mid-level.

    The hourly hydropower (MW) is calculated using the following formula:

    hourly_hydropower = min(turbine_factor * rho * g * head * flow_rate, max_generation)

    where:
    * rho: Density of water (kg/m3)
    * g: Acceleration due to gravity (m/s2)
    * head: Water head above the turbine (m)
    * flow_rate: Water flow rate (m3/hour)
    * max_generation: Maximum power generation capacity of the turbine (MW)
    * turbine_factor: Turbine efficiency

    Args:
        release: Water release (m3/hour)
        mid_level: Average water level between current and previous timestep (m)
        max_generation: Maximum power generation capacity of the turbine (MW)
        turbine_factor: Turbine efficiency
        max_head: Maximum head (m)
        max_level: Maximum water level (m)

    Returns:
        pd.Series: Hourly hydropower generation (MW)
    """
    # Define constants
    density = 998  # kg/m3
    gravity = 9.81  # m/s2

    # Calculate the water head above the turbine
    head = max_head - (max_level - mid_level)
    # Convert release from m3/hour to m3/s because the hydropower formula uses seconds
    flow_rate = release / 3600

    # The calculated hourly hydropower is in Watts
    hourly_hydropower = turbine_factor * density * gravity * head * flow_rate
    hourly_hydropower = hourly_hydropower / 1e6  # Convert to MegaWatts
    # A turbine has a maximum water intake, so the power generation is capped
    hourly_hydropower = np.minimum(hourly_hydropower, max_generation)
    return hourly_hydropower


def calc_daily_hydropower(
    release: pd.Series,
    mid_level: pd.Series,
    max_generation: float,
    turbine_factor: float,
    max_head: float,
    max_level: float,
) -> pd.Series:
    """Calculate daily hydropower generation from release and mid-level.

    Args:
        release: Water release (m3/daily)
        mid_level: Average water level between current and previous timestep (m)
        max_generation: Maximum power generation capacity of the turbine (MW)
        turbine_factor: Turbine efficiency
        max_head: Maximum head (m)
        max_level: Maximum water level (m)

    Returns:
        pd.Series: Daily hydropower generation (MW-day)
    """
    # Convert release from m3/daily to m3/hour
    release = release / 24  # Convert to m3/hour
    hourly_hydropower = calc_hourly_hydropower(
        release=release,
        mid_level=mid_level,
        max_generation=max_generation,
        turbine_factor=turbine_factor,
        max_head=max_head,
        max_level=max_level,
    )
    # Convert to MW-day
    daily_hydropower = hourly_hydropower * 24
    return daily_hydropower


def calc_release_impact(
    release_t: float,
    storage_t0: float,
    total_inflow_t: float,
    min_level: float,
    max_level: float,
    max_storage: float,
    level_t0: float,
    max_generation: float,
    turbine_factor: float,
    max_head: float,
) -> tuple[float, float, float, float]:
    """Calculate the impact of a given release on storage, level, mid-level, and hydropower."""
    spill_t = max(
        storage_t0 + total_inflow_t - max_storage - release_t,
        0,
    )
    storage_t = storage_t0 + total_inflow_t - release_t - spill_t
    level_t = calc_level_from_storage(
        storage=storage_t,
        min_level=min_level,
        max_level=max_level,
        max_storage=max_storage,
    )
    mid_level_t = (level_t0 + level_t) / 2
    daily_hydropower_t = calc_daily_hydropower(
        release=release_t,
        mid_level=mid_level_t,
        max_generation=max_generation,
        turbine_factor=turbine_factor,
        max_head=max_head,
        max_level=max_level,
    )
    return spill_t, storage_t, level_t, daily_hydropower_t


def calc_max_release(
    total_inflow_t: float,
    release_t0: float,
    storage_t0: float,
    min_flow_t: float,
    max_release: float,
    hydropeak_factor: float,
) -> float:
    """Calculate the maximum allowable release of the current timestep."""
    # Release is limited by the hydropeaking factor
    max_release_hydropeak_t = release_t0 + max_release * hydropeak_factor
    # Release cannot be larger than the turbine capacity
    max_release_t = min(max_release, max_release_hydropeak_t)
    # Release cannot be less than the min environmental flow
    max_release_t = max(min_flow_t, max_release_t)
    # Cannot release more than the amount of water in the reservoir
    if storage_t0 + total_inflow_t - max_release_t < 0:
        max_release_t = storage_t0 + total_inflow_t
    return max_release_t


def calc_min_release(
    total_inflow_t: float,
    release_t0: float,
    storage_t0: float,
    min_flow_t: float,
    max_release: float,
    hydropeak_factor: float,
) -> float:
    """Calculate the minimum allowable release of the current timestep."""
    # Release is limited by the hydropeaking factor
    min_release_hydropeak_t = release_t0 - max_release * hydropeak_factor
    # Limited by the minimum environmental flow
    min_release_t = max(min_flow_t, min_release_hydropeak_t)
    # Release cannot make the storage become negative
    if storage_t0 + total_inflow_t - min_release_t < 0:
        min_release_t = storage_t0 + total_inflow_t
    return min_release_t


def convert_to_hourly_hydropower(daily_hydropower) -> pd.Series:
    """Return the hourly hydropower values."""
    hourly_hydropower = daily_hydropower / 24
    # Repeat the hydropower values for each hour of the day
    hourly_hydropower = hourly_hydropower.loc[
        hourly_hydropower.index.repeat(24)
    ].reset_index(drop=True)
    # Index starts from 1
    hourly_hydropower.index += 1
    return hourly_hydropower



""" data_processor.py: This file contains the DataProcessor class that processes the data provided by the user.
"""

import json
import os

import networkx as nx
import numpy as np
import pandas as pd

from pownet.folder_utils import get_database_dir
from pownet.data_utils import get_dates


class DataProcessor:
    def __init__(
        self, input_folder: str, model_name: str, year: int, frequency: int
    ) -> None:
        """The DataProcessor class is used to process the data provided by the user. The data
        is stored in the model_library/model_name folder. The required files are:
        1. transmission.csv: A file that contains the transmission data.
        2. thermal_unit.csv: A file that contains the thermal unit data.
        3. solar.csv, wind.csv, hydropower.csv, import.csv: Files that contain the renewable unit data.
        4. energy_storage.csv: A file that contains the energy storage system data.
        """
        self.input_folder = input_folder
        self.model_name = model_name
        self.year = year
        self.frequency = frequency

        # Values that will be calculated
        self.cycle_map: dict = json.loads("{}")
        self.thermal_derate_factors: pd.DataFrame = pd.DataFrame()
        self.thermal_derated_capacity: pd.DataFrame = pd.DataFrame()

        self.transmission_data: pd.DataFrame = pd.DataFrame()
        self.transmission_params: dict = {}  # Default PowNet parameters
        self.user_transmission: pd.DataFrame = pd.DataFrame()

        self.ess_derate_factors: pd.DataFrame = pd.DataFrame()
        self.ess_derated_capacity: pd.DataFrame = pd.DataFrame()

        # Maps frequency to wavelength
        wavelengths = {50: 6000, 60: 5000}
        self.wavelength = wavelengths[frequency]

        # Note that we will modify the original file
        self.model_folder = os.path.join(self.input_folder, model_name)

    def load_transmission_data(self) -> None:
        # User inputs of transmission data
        self.user_transmission = pd.read_csv(
            os.path.join(self.model_folder, "transmission.csv"),
            header=0,
        )

        # First check that there are no repeated edges in the transmission data
        # Here, only one direction is needed. If the reverse direction is also
        # present, then it is a repeated edge.
        edges = self.user_transmission[["source", "sink"]].values
        reversed_edges = np.flip(edges, axis=1)
        edges_set = set(map(tuple, edges))
        reversed_edges_set = set(map(tuple, reversed_edges))
        repeated_edges = edges_set.intersection(reversed_edges_set)
        if len(repeated_edges) > 0:
            raise ValueError(
                f"There are repeated edges in the transmission data: {repeated_edges}"
            )

        # Generic transmission parameters for a power system
        self.transmission_params: dict = (
            pd.read_csv(
                os.path.join(get_database_dir(), "transmission_params.csv"), header=0
            )
            .set_index("kv")
            .to_dict(orient="dict")
        )
        # Calculated transmission data
        self.transmission_data: pd.DataFrame = self.user_transmission[
            ["source", "sink", "source_kv", "sink_kv"]
        ].copy()

    def calc_stability_limit(
        self,
        source_kv: int,
        sink_kv: int,
        distance: float,
        wavelength: int,
        n_circuits: int,
    ) -> float:
        """This function calculates the steady-state stability limit of a transmission line.
        From Chapter 5 of Power System Analysis and Design 5th (EQ 5.4.30)

        The stability limit per circuit is given by:
                P = V1 * V2 / X * sin(2 * pi * d / lambda)
            where:
                P is the stability limit in MW
                V1 and V2 are the voltages at the two ends of the line
                X is the reactance of the line in ohms per km
                d is the distance between the two ends of the line in km
                lambda is the wavelength of the system in km

        Args:
            source_kv (int): Voltage level of the source bus
            sink_kv (int): Voltage level of the sink bus
            distance (float): Distance between the two buses in km
            wavelength (int): Wavelength of the system in km
            n_circuits (int): Number of circuits in the transmission line

        Returns:
            float: The stability limit of the transmission line in MW
        """
        # The reactance of the line is a function of the maximum voltage level
        # of the two buses.
        max_kv = max(source_kv, sink_kv)
        reactance_per_km = self.transmission_params["reactance_ohms_per_km"][max_kv]
        # Calculate the Surge Impedance Limit (SIL)
        sil = source_kv * sink_kv / reactance_per_km / 1000  # Divide by 1000 to get MW
        stability_limit_per_circuit = sil / np.sin(2 * np.pi * distance / wavelength)
        return int(n_circuits * stability_limit_per_circuit)

    def calc_thermal_limit(
        self, source_kv: int, sink_kv: int, n_circuits: int
    ) -> float:
        """From Chapter 5 of Power System Analysis and Design 5th. See Example 5.6b.
        The full-load current at 1 per-unit factor is
              I = P/(sqrt(3) * V)
        Here, P is the surge impedance factor (SIL) and V is the voltage of the
        receiving bus. This voltage is the minimum voltage between the two ends.

        Args:
            source_kv (int): Voltage level of the source bus
            sink_kv (int): Voltage level of the sink bus
            n_circuits (int): Number of circuits in the transmission line

        Returns:
            float: The thermal limit of the transmission line in MW
        """
        max_kv = max(source_kv, sink_kv)
        n_conductors = self.transmission_params["n_conductors"][max_kv]
        # in Amps
        current_capacity = self.transmission_params["current_capacity_amps"][max_kv]
        total_current_capacity = n_conductors * current_capacity / 1000  # in kilo-A
        thermal_limit_per_circuit = total_current_capacity * np.sqrt(3) * max_kv
        return int(n_circuits * thermal_limit_per_circuit)

    def calc_line_capacity(self) -> None:
        """Calculate the capacity of line segments. The unit is in MW.
        Line capacity is the minimum of the thermal limit and the steady-state
        stability limit (a function of distance).

        Note the calculated values are overwritten by user provided values
        in the transmission.csv file.
        """
        self.transmission_data["stability_limit"] = self.user_transmission.apply(
            lambda x: self.calc_stability_limit(
                x["source_kv"],
                x["sink_kv"],
                x["distance"],
                self.wavelength,
                x["n_circuits"],
            ),
            axis=1,
        )

        self.transmission_data["thermal_limit"] = self.user_transmission.apply(
            lambda x: self.calc_thermal_limit(
                x["source_kv"], x["sink_kv"], x["n_circuits"]
            ),
            axis=1,
        )
        # The transmission limit is the minimum of the thermal limit and
        # the steady-state steability limit (a function of distance).
        # Normally, distance is the deciding factor.
        self.transmission_data["line_capacity"] = self.transmission_data[
            ["thermal_limit", "stability_limit"]
        ].min(axis=1)

        # Overwrite calculated values with user provided values
        excluded_list = [-1, None]
        user_specified_capacity = self.user_transmission.loc[
            ~self.user_transmission["user_line_cap"].isin(excluded_list)
        ]
        user_specified_capacity = user_specified_capacity.set_index(
            ["source", "sink"]
        ).rename(columns={"user_line_cap": "line_capacity"})

        self.transmission_data = self.transmission_data.set_index(["source", "sink"])
        self.transmission_data.update(user_specified_capacity)
        self.transmission_data = self.transmission_data.reset_index()

    def calc_line_susceptance(self) -> None:
        """Calculate the susceptance of line segments. The unit is in Siemens (S)."""
        # Assume reactance based on the maximum voltage level of the two buses
        self.transmission_data["max_kv"] = self.user_transmission.apply(
            lambda x: max(x["source_kv"], x["sink_kv"]), axis=1
        )

        self.transmission_data["reactance_per_km"] = self.transmission_data.apply(
            lambda x: self.transmission_params["reactance_ohms_per_km"][x["max_kv"]],
            axis=1,
        )

        self.transmission_data["reactance"] = (
            self.transmission_data["reactance_per_km"]
            * self.user_transmission["distance"]
        )

        self.transmission_data["susceptance"] = self.transmission_data.apply(
            lambda x: int(x["source_kv"] * x["sink_kv"] / x["reactance"]),
            axis=1,
        )

        # Replace with user-specified values
        excluded_values = [-1, None]
        user_specified_susceptance = self.user_transmission.loc[
            ~self.user_transmission["user_susceptance"].isin(excluded_values),
            ["source", "sink", "user_susceptance"],
        ]
        user_specified_susceptance = user_specified_susceptance.set_index(
            ["source", "sink"]
        )
        # Change from float to int
        user_specified_susceptance = user_specified_susceptance.astype(
            {"user_susceptance": int}
        )

        self.transmission_data = self.transmission_data.set_index(["source", "sink"])
        self.transmission_data.update(user_specified_susceptance)
        self.transmission_data = self.transmission_data.reset_index()

    def write_transmission_data(self) -> None:
        self.transmission_data.to_csv(
            os.path.join(self.model_folder, "pownet_transmission.csv"), index=False
        )

    def create_cycle_map(self) -> None:
        """
        Create a cycle map for the power system. This is used to create the
        cycle constraints in the optimization model. The cycle map is a dictionary
        where the key is the cycle name and the value is a list of nodes in the cycle.
        """
        graph = nx.from_pandas_edgelist(
            self.transmission_data,
            source="source",
            target="sink",
        )
        cycles = nx.cycle_basis(graph)
        # Save this map to be uses by ModelBuilder
        self.cycle_map = {f"cycle_{idx+1}": cycle for idx, cycle in enumerate(cycles)}

    def write_cycle_map(self) -> None:
        """
        Save the cycle map to a json file in model_library/{model_name}.
        The key is the cycle name and the value is a list of nodes in the cycle.
        """
        with open(os.path.join(self.model_folder, "pownet_cycle_map.json"), "w") as f:
            json.dump(self.cycle_map, f)

    def _create_derate_factors(
        self, unit_type: str, derate_factor: float = 1.00
    ) -> None:
        """Creates derate factors for a given unit type (thermal or ess).

        Args:
            unit_type (str): The type of unit ('thermal' or 'ess').
            derate_factor (float): The derate factor to apply. Defaults to 1.00.
        """

        model_dir = os.path.join(self.input_folder, self.model_name)

        if unit_type == "thermal":
            filename = "thermal_unit.csv"
            attribute_name = "thermal_derate_factors"
        elif unit_type == "ess":
            filename = "energy_storage.csv"
            attribute_name = "ess_derate_factors"
        else:
            raise ValueError(
                f"Invalid unit type: {unit_type}. Must be 'thermal' or 'ess'."
            )

        if os.path.exists(os.path.join(model_dir, filename)):
            units = pd.read_csv(os.path.join(model_dir, filename))["name"].values
        else:
            return

        num_hrs_in_year = 8760
        temp_df = pd.DataFrame(
            derate_factor,
            index=range(0, num_hrs_in_year),
            columns=units,
        )
        setattr(
            self,
            attribute_name,
            pd.concat([get_dates(year=self.year), temp_df], axis=1),
        )

    def create_thermal_derate_factors(self, derate_factor: float = 1.00) -> None:
        """Creates derate factors for thermal units."""
        self._create_derate_factors("thermal", derate_factor)

    def create_ess_derate_factors(self, derate_factor: float = 1.00) -> None:
        """Creates derate factors for ESS units."""
        self._create_derate_factors("ess", derate_factor)

    def write_thermal_derate_factors(self) -> None:
        self.thermal_derate_factors.to_csv(
            os.path.join(self.model_folder, "pownet_derate_factor.csv"), index=False
        )

    def _create_derated_capacity(self, unit_type: str) -> None:
        """Creates a dataframe of hourly derated capacity for a given unit type.

        Args:
            unit_type (str): The type of unit ('thermal' or 'ess').
        """

        if unit_type == "thermal":
            filename = "thermal_unit.csv"
            derate_factor_attr = "thermal_derate_factors"
            derated_capacity_attr = "thermal_derated_capacity"
        elif unit_type == "ess":
            filename = "energy_storage.csv"
            derate_factor_attr = "ess_derate_factors"
            derated_capacity_attr = "ess_derated_capacity"
        else:
            raise ValueError(
                f"Invalid unit type: {unit_type}. Must be 'thermal' or 'ess'."
            )

        # Get the nameplate capacity of each unit
        filepath = os.path.join(self.model_folder, filename)
        if os.path.exists(filepath):
            max_cap = pd.read_csv(
                filepath,
                index_col="name",
                usecols=["name", "max_capacity"],
            )[
                "max_capacity"
            ]  # Directly get the Series
        else:
            return

        # Get the derate factors for the units
        derate_factors = getattr(self, derate_factor_attr)

        # Efficiently calculate derated capacity using vectorized operations
        derated_capacity = derate_factors.drop(columns=["date", "hour"]).mul(
            max_cap, axis=1
        )

        # Concatenate with dates and set the index
        derated_capacity = pd.concat(
            [get_dates(year=self.year), derated_capacity], axis=1
        )
        derated_capacity.index += 1

        setattr(self, derated_capacity_attr, derated_capacity)

    def create_thermal_derated_capacity(self) -> None:
        """Creates a dataframe of hourly derated capacity of thermal units."""
        self._create_derated_capacity("thermal")

    def create_ess_derated_capacity(self) -> None:
        """Creates a dataframe of hourly derated capacity of ess units."""
        self._create_derated_capacity("ess")

    def write_thermal_derated_capacity(self) -> None:
        if not self.thermal_derate_factors.empty:
            self.thermal_derated_capacity.to_csv(
                os.path.join(self.model_folder, "pownet_thermal_derated_capacity.csv"),
                index=False,
            )

    def write_ess_derated_capacity(self) -> None:
        if not self.ess_derated_capacity.empty:
            self.ess_derated_capacity.to_csv(
                os.path.join(self.model_folder, "pownet_ess_derated_capacity.csv"),
                index=False,
            )

    def check_user_line_capacities(self) -> None:
        """The user can provide their own line capacities under user_line_cap column
        in transmission.csv. If this is the case, then it will be used instead of the
        calculated line capacities.
        """
        pass

    def run_all_processing_steps(self) -> None:
        """Run all the data processing steps"""
        if not self.user_transmission.empty:
            self.calc_line_capacity()
            self.calc_line_susceptance()
            self.create_cycle_map()

        self.create_thermal_derate_factors()
        self.create_thermal_derated_capacity()

        self.create_ess_derate_factors()
        self.create_ess_derated_capacity()

    def write_data(self) -> None:
        """Write the processed data as csv files sharing a prefix "pownet_" to the model folder"""

        if not self.transmission_data.empty:
            self.write_transmission_data()
            self.write_cycle_map()

        self.write_thermal_derated_capacity()
        self.write_ess_derated_capacity()

    def execute_data_pipeline(self) -> None:
        if os.path.exists(os.path.join(self.model_folder, "transmission.csv")):
            self.load_transmission_data()

        self.run_all_processing_steps()
        self.write_data()



from datetime import datetime
from pownet.core import (
    DataProcessor,
    SystemInput,
    ModelBuilder,
    SystemRecord,
    OutputProcessor,
    Visualizer,
)
from pownet.data_utils import create_init_condition

##### User inputs #####
to_process_inputs = True
sim_horizon = 24
model_name = "dummy"
steps_to_run = 3  # Default is None
do_plot = True
#######################

if to_process_inputs:
    data_processor = DataProcessor(
        input_folder="../model_library", model_name=model_name, year=2016, frequency=50
    )
    data_processor.execute_data_pipeline()

inputs = SystemInput(
    input_folder="../model_library",
    model_name=model_name,
    year=2016,
    sim_horizon=sim_horizon,
    spin_reserve_factor=0.15,
    load_shortfall_penalty_factor=1000,
    load_curtail_penalty_factor=1,
    spin_shortfall_penalty_factor=1000,
)
inputs.load_and_check_data()


model_builder = ModelBuilder(inputs)
record = SystemRecord(inputs)

build_times = []
opt_times = []
objvals = []
init_conditions = create_init_condition(inputs.thermal_units)

if steps_to_run is None:
    steps_to_run = 10  # 365 - (sim_horizon // 24 - 1)

for step_k in range(1, steps_to_run):
    start_time = datetime.now()
    if step_k == 1:
        power_system_model = model_builder.build(
            step_k=step_k,
            init_conds=init_conditions,
        )
    else:
        power_system_model = model_builder.update(
            step_k=step_k,
            init_conds=init_conditions,
        )
    build_times.append((datetime.now() - start_time).total_seconds())

    power_system_model.optimize(mipgap=0.001)

    # Raise an error if the model is not feasible
    if not power_system_model.check_feasible():
        raise ValueError("Model is not feasible.")

    objvals.append(power_system_model.get_objval())
    opt_times.append(power_system_model.get_runtime())

    record.keep(
        runtime=power_system_model.get_runtime(),
        objval=power_system_model.get_objval(),
        solution=power_system_model.get_solution(),
        lmp=power_system_model.solve_for_lmp(),
        step_k=step_k,
    )
    init_conditions = record.get_init_conds()


# Process the results
output_processor = OutputProcessor(inputs)
node_var_df = record.get_node_variables()
output_processor.load_from_dataframe(node_var_df)

# Visualize the results
if do_plot:
    visualizer = Visualizer(inputs.model_id)
    if steps_to_run <= 3:
        visualizer.plot_fuelmix_bar(
            dispatch=output_processor.get_hourly_dispatch(),
            demand=output_processor.get_hourly_demand(),
        )
    else:
        visualizer = Visualizer(inputs.model_id)
        visualizer.plot_fuelmix_area(
            dispatch=output_processor.get_daily_dispatch(),
            demand=output_processor.get_daily_demand(),
        )

# # Save other modeling statistics
# import os
# from pownet.folder_utils import get_output_dir
# import pandas as pd

# folder_dir = get_output_dir()
# prefix_name = "warmstart"
# df = pd.DataFrame(
#     {
#         "build_time": build_times,
#         "opt_time": opt_times,
#         "objval": objvals,
#     }
# )
# df.to_csv(
#     os.path.join(
#         folder_dir,
#         f"branch{prefix_name}_{inputs.model_name}_{sim_horizon}_modelstats.csv",
#     ),
#     index=False,
# )



"""constraint.py: Contains functions for constructing the objective function and constraints of the optimization model.
Throughout this module, the number 24 is the number of hours in a day. This number is used to calculate the index of the next day in the optimization model."""

import gurobipy as gp
import networkx as nx
import pandas as pd


def get_capacity_value(t: int, unit: str, step_k: int, capacity_df) -> float:
    value = capacity_df.loc[t + (step_k - 1) * 24, unit]
    if isinstance(value, pd.Series):
        return value.iloc[0]
    return value


"""Functions for thermal-unit constraints
"""


def add_c_link_uvw_init(
    model: gp.Model,
    u: gp.tupledict,
    v: gp.tupledict,
    w: gp.tupledict,
    thermal_units: list,
    initial_u: dict,
) -> gp.tupledict:
    """Equation 2 of Kneuven et al (2019) based on Garver (1962).
    Three binary variables u (status), v (startup), and w (shutdown) are used to model the status of a thermal unit.
    Their logical relationship is defined by this equation. The initial condition is also considered here.

    Args:
        model (gp.Model): The optimization model
        u (gp.tupledict): The status of the thermal unit
        v (gp.tupledict): The startup of the thermal unit
        w (gp.tupledict): The shutdown of the thermal unit
        thermal_units (list): The list of thermal units
        initial_u (dict): The initial status of the thermal unit

    Returns:
        gp.tupledict: The constraints for the initial condition

    """
    return model.addConstrs(
        (
            u[unit_g, 1]
            # Last hour of the previous iteration
            - initial_u[unit_g] == v[unit_g, 1] - w[unit_g, 1]
            for unit_g in thermal_units
        ),
        name="link_uvw_init",
    )


def add_c_link_uvw(
    model: gp.Model,
    u: gp.tupledict,
    v: gp.tupledict,
    w: gp.tupledict,
    sim_horizon: int,
    thermal_units: list,
) -> gp.tupledict:
    """Equation 2 of Kneuven et al (2019) based on Garver (1962).
    Three binary variables u (status), v (startup), and w (shutdown) are used to model the status of a thermal unit.
    Their logical relationship is defined by this equation.

    Args:
        model (gp.Model): The optimization model
        u (gp.tupledict): The status of the thermal unit
        v (gp.tupledict): The startup of the thermal unit
        w (gp.tupledict): The shutdown of the thermal unit
        sim_horizon (int): The simulation horizon
        thermal_units (list): The list of thermal units

    Returns:
        gp.tupledict: The constraints for the subsequent time

    """
    # Only consider the second timestep onwards
    return model.addConstrs(
        (
            u[unit_g, t] - u[unit_g, t - 1] == v[unit_g, t] - w[unit_g, t]
            for unit_g in thermal_units
            for t in range(2, sim_horizon + 1)
        ),
        name="link_uvw",
    )


def add_c_link_pthermal(
    model: gp.Model,
    pthermal: gp.tupledict,
    p: gp.tupledict,
    u: gp.tupledict,
    timesteps: range,
    thermal_units: list,
    thermal_min_capacity: dict,
) -> gp.tupledict:
    """Equation 12 of Kneuven et al (2019). This establishes the power output above
    the minimum capacity, which is a variable promoted by Morales-Espana et al. (2013).

    Args:
        model (gp.Model): The optimization model
        pthermal (gp.tupledict): The power output
        p (gp.tupledict): The power output above the minimum capacity
        u (gp.tupledict): The status of the thermal unit
        timesteps (range): The range of timesteps
        thermal_units (list): The list of thermal units
        thermal_min_capacity (dict): The minimum capacity of the thermal unit

    Returns:
        gp.tupledict: The constraints for the power output above the minimum capacity

    """
    return model.addConstrs(
        (
            pthermal[unit_g, t]
            == p[unit_g, t] + thermal_min_capacity[unit_g] * u[unit_g, t]
            for unit_g in thermal_units
            for t in timesteps
        ),
        name="link_pthermal",
    )


def add_c_link_spin(
    model: gp.Model,
    p: gp.tupledict,
    pbar: gp.tupledict,
    spin: gp.tupledict,
    timesteps: range,
    thermal_units,
) -> gp.tupledict:
    """Equation 14 of Kneuven et al (2019). Establish the relationship among p, pbar, and spin:
    pbar = p + spin
    Note that pbar and p are variables defines as MW above the minimum capacity.

    Args:
        model (gp.Model): The optimization model
        p (gp.tupledict): The power output above the minimum capacity
        pbar (gp.tupledict): The maximum power output above the minimum capacity
        spin (gp.tupledict): The spinning reserve
        timesteps (range): The range of timesteps
        thermal_units (list): The list of thermal units

    Returns:
        gp.tupledict: The constraints for the power output above the minimum capacity

    """
    return model.addConstrs(
        (
            pbar[unit_g, t] == p[unit_g, t] + spin[unit_g, t]
            for unit_g in thermal_units
            for t in timesteps
        ),
        name="link_spin",
    )


def add_c_link_ppbar(
    model: gp.Model,
    p: gp.tupledict,
    pbar: gp.tupledict,
    timesteps: range,
    thermal_units: list,
) -> gp.tupledict:
    """
    Equation 17 of Kneuven et al (2019). Since pbar = p + spin, if we do not implement
    the spin variable, then p <= pbar.
    """
    return model.addConstrs(
        (
            pbar[unit_g, t] >= p[unit_g, t]
            for t in timesteps
            for unit_g in thermal_units
        ),
        name="link_ppbar",
    )


def add_c_link_pu_lower(
    model: gp.Model,
    pthermal: gp.tupledict,
    u: gp.tupledict,
    timesteps: range,
    thermal_units: list,
    thermal_min_capacity: dict,
) -> gp.tupledict:
    """Equation 18 of Kneuven et al (2019) based on Carrion and Arroyo (2006).
    Set the lower bound of the dispatched power.

    Args:
        model (gp.Model): The optimization model
        pthermal (gp.tupledict): The dispatched power
        u (gp.tupledict): The status of the thermal unit
        timesteps (range): The range of timesteps
        thermal_units (list): The list of thermal units
        thermal_min_capacity (dict): The minimum capacity of the thermal unit

    Returns:
        gp.tupledict: The constraints for the lower bound of the dispatched power
    """
    return model.addConstrs(
        (
            thermal_min_capacity[unit_g] * u[unit_g, t] <= pthermal[unit_g, t]
            for unit_g in thermal_units
            for t in timesteps
        ),
        name="pthermal_lb",
    )


def add_c_link_pu_upper(
    model=gp.Model,
    pbar=gp.tupledict,
    u=gp.tupledict,
    timesteps=range,
    step_k=int,
    thermal_units=list,
    thermal_min_capacity=dict,
    thermal_derated_capacity=pd.DataFrame,
) -> gp.tupledict:
    """Equation 18 of Knueven et al (2019) based on Carrion and Arroyo (2006).
    Set the upper bound of the dispatched power.

    Args:
        model (gp.Model): The optimization model
        pbar (gp.tupledict): The maximum power output above the minimum capacity
        u (gp.tupledict): The status of the thermal unit
        timesteps (range): The range of timesteps
        step_k (int): The current iteration
        thermal_units (list): The list of thermal units
        thermal_min_capacity (dict): The minimum capacity of the thermal unit
        thermal_derated_capacity (pd.DataFrame): The derated capacity of the thermal unit

    Returns:
        gp.tupledict: The constraints for the upper bound of the dispatched power

    """
    return model.addConstrs(
        (
            pbar[unit, t] + thermal_min_capacity[unit] * u[unit, t]
            <= thermal_derated_capacity.loc[t + (step_k - 1) * 24, unit] * u[unit, t]
            for unit in thermal_units
            for t in timesteps
        ),
        name="pthermal_ub",
    )


def add_c_thermal_curtail(
    model: gp.Model,
    pthermal: gp.tupledict,
    pthermal_curtail: gp.tupledict,
    pcharge: gp.tupledict,
    timesteps: range,
    step_k: int,
    thermal_must_take_units: list,
    thermal_derated_capacity: pd.DataFrame,
    ess_attached: dict,
) -> gp.tupledict:
    """For must-take thermal units, the curtailment is the difference between the derated capacity and the dispatched power."""
    constraints = gp.tupledict()
    for unit in thermal_must_take_units:
        has_storage = unit in ess_attached
        for t in timesteps:
            pcharge_unit_t = 0
            if has_storage:
                # A unit may have multiple storage systems
                for storage_unit in ess_attached[unit]:
                    pcharge_unit_t += pcharge[storage_unit, t]

            cname = f"thermal_curtail[{unit},{t}]"
            constraints[cname] = model.addConstr(
                (
                    pthermal[unit, t] + pthermal_curtail[unit, t] + pcharge_unit_t
                    == thermal_derated_capacity.loc[t + (step_k - 1) * 24, unit]
                ),
                name=cname,
            )
    return constraints


def add_c_min_down_init(
    model: gp.Model,
    u: gp.tupledict,
    sim_horizon: int,
    thermal_units: list,
    initial_min_off: dict,
) -> gp.tupledict:
    """Equation 3b of Kneuven et al (2019). Minimum downtime of thermal units at t=1.

    Args:
        model (gp.Model): The optimization model
        u (gp.tupledict): The status of the thermal unit
        sim_horizon (int): The simulation horizon
        thermal_units (list): The list of thermal units
        initial_min_off (dict): The remaining minimum downtime of the thermal unit from the previous iteration

    Returns:
        gp.tupledict: The constraints for the minimum downtime of thermal units at t=1

    """
    constraints = gp.tupledict()
    for unit_g in thermal_units:
        cname = f"minDownInit[{unit_g}]"
        # Find the min between the required downtime and the simulation horizon
        min_DT = min(initial_min_off[unit_g], sim_horizon)
        constraints[cname] = model.addConstr(
            u.sum(unit_g, range(1, min_DT + 1)) == 0,
            name=cname,
        )
    return constraints


def add_c_min_up_init(
    model: gp.Model,
    u: gp.tupledict,
    sim_horizon: int,
    thermal_units: list,
    initial_min_on: dict,
) -> gp.tupledict:
    """Equation 3a of Kneuven et al (2019). Minimum uptime of thermal units at t=1.

    Args:
        model (gp.Model): The optimization model
        u (gp.tupledict): The status of the thermal unit
        sim_horizon (int): The simulation horizon
        thermal_units (list): The list of thermal units
        initial_min_on (dict): The remaining minimum uptime of the thermal unit from the previous iteration

    Returns:
        gp.tupledict: The constraints for the minimum uptime of thermal units at t=1

    """
    constraints = gp.tupledict()
    for unit_g in thermal_units:
        cname = f"minUpInit[{unit_g}]"
        # Find the min between the required uptime and the simulation horizon
        min_UT = min(initial_min_on[unit_g], sim_horizon)
        constraints[cname] = model.addConstr(
            u.sum(unit_g, range(1, min_UT + 1)) == min_UT,
            name=cname,
        )
    return constraints


def add_c_min_down(
    model: gp.Model,
    u: gp.tupledict,
    w: gp.tupledict,
    sim_horizon: int,
    thermal_units: list,
    TD: dict,
) -> gp.tupledict:
    """Equation 5 of Kneuven et al (2019) based on Malkin (2003) and Rajan and Takriti (2005).
    Minimum downtime of thermal units at t>1.

    Args:
        model (gp.Model): The optimization model
        u (gp.tupledict): The status of the thermal unit
        w (gp.tupledict): The shutdown of the thermal unit
        sim_horizon (int): The simulation horizon
        thermal_units (list): The list of thermal units
        TD (dict): The minimum downtime of thermal units

    Returns:
        gp.tupledict: The constraints for the minimum downtime of thermal units at t>1

    """
    constraints = gp.tupledict()
    for unit_g in thermal_units:
        TD_g = TD[unit_g]
        for t in range(TD_g, sim_horizon + 1):
            cname = f"minDown[{unit_g},{t}]"
            LHS = gp.quicksum([w[unit_g, i] for i in range(t - TD_g + 1, t + 1)])
            constraints[cname] = model.addConstr(LHS <= 1 - u[unit_g, t], name=cname)
    return constraints


def add_c_min_up(
    model=gp.Model,
    u=gp.tupledict,
    v=gp.tupledict,
    sim_horizon=int,
    thermal_units=list,
    TU=dict,
) -> gp.tupledict:
    """Equation 4 of Kneuven et al (2019) based on Malkin (2003) and Rajan and Takriti (2005).
    Minimum uptime of thermal units at t>1.

    Args:
        model (gp.Model): The optimization model
        u (gp.tupledict): The status of the thermal unit
        v (gp.tupledict): The startup of the thermal unit
        sim_horizon (int): The simulation horizon
        thermal_units (list): The list of thermal units
        TU (dict): The minimum uptime of thermal units

    Returns:
        gp.tupledict: The constraints for the minimum uptime of thermal units at t>1

    """
    constraints = gp.tupledict()
    for unit_g in thermal_units:
        TU_g = TU[unit_g]
        for t in range(TU_g, sim_horizon + 1):
            cname = f"minUp[{unit_g},{t}]"
            LHS = gp.quicksum([v[unit_g, i] for i in range(t - TU_g + 1, t + 1)])
            constraints[cname] = model.addConstr(LHS <= u[unit_g, t], name=cname)
    return constraints


def add_c_peak_down_bound(
    model: gp.Model,
    p: gp.tupledict,
    spin: gp.tupledict,
    u: gp.tupledict,
    v: gp.tupledict,
    w: gp.tupledict,
    sim_horizon: int,
    step_k: int,
    thermal_units: list,
    thermal_min_capacity: dict,
    thermal_derated_capacity: pd.DataFrame,
    SD: dict,
    SU: dict,
    TU: dict,
) -> gp.tupledict:
    """Equation 23b of Kneuven et al (2019) based on Gentile et al. (2017).
    When a peaking thermal unit (TD=1) is shutting down, its power output is limited by
    the shutdown rate (SD).

    Args:
        model (gp.Model): The optimization model
        p (gp.tupledict): The power output above the minimum capacity
        spin (gp.tupledict): The spinning reserve
        u (gp.tupledict): The status of the thermal unit
        v (gp.tupledict): The startup of the thermal unit
        w (gp.tupledict): The shutdown of the thermal unit
        sim_horizon (int): The simulation horizon
        step_k (int): The current iteration
        thermal_units (list): The list of thermal units
        thermal_min_capacity (dict): The minimum capacity of the thermal unit
        thermal_derated_capacity (pd.DataFrame): The derated capacity of the thermal unit
        SD (dict): The shutdown rate of the thermal unit
        SU (dict): The startup rate of the thermal unit
        TU (dict): The minimum uptime of thermal units

    Returns:
        gp.tupledict: The constraints for the peak down bound

    """
    return model.addConstrs(
        (
            p[unit_g, t] + spin[unit_g, t]
            <= (
                thermal_derated_capacity.loc[t + (step_k - 1) * 24, unit_g]
                - thermal_min_capacity[unit_g]
            )
            * u[unit_g, t]
            - (thermal_derated_capacity.loc[t + (step_k - 1) * 24, unit_g] - SD[unit_g])
            * w[unit_g, t + 1]
            - max(0, (SD[unit_g] - SU[unit_g])) * v[unit_g, t]
            for t in range(
                1, sim_horizon
            )  # The constraints index w at t+1, so we stop at T-1
            for unit_g in thermal_units
            if TU[unit_g] == 1
        ),
        name="peakDownBnd",
    )


def add_c_peak_up_bound(
    model: gp.Model,
    p: gp.tupledict,
    spin: gp.tupledict,
    u: gp.tupledict,
    v: gp.tupledict,
    w: gp.tupledict,
    sim_horizon: int,
    step_k: int,
    thermal_units: list,
    thermal_min_capacity: dict,
    thermal_derated_capacity: pd.DataFrame,
    SD: dict,
    SU: dict,
    TU: dict,
) -> gp.tupledict:
    """Equation 23a of Kneuven et al (2019) based on Gentile et al. (2017).
    When a peaking thermal unit (TU=1) is starting up, its power output is limited by
    the startup rate (SU).

    Args:
        model (gp.Model): The optimization model
        p (gp.tupledict): The power output above the minimum capacity
        spin (gp.tupledict): The spinning reserve
        u (gp.tupledict): The status of the thermal unit
        v (gp.tupledict): The startup of the thermal unit
        w (gp.tupledict): The shutdown of the thermal unit
        sim_horizon (int): The simulation horizon
        step_k (int): The current iteration
        thermal_units (list): The list of thermal units
        thermal_min_capacity (dict): The minimum capacity of the thermal unit
        thermal_derated_capacity (pd.DataFrame): The derated capacity of the thermal unit
        SD (dict): The shutdown rate of the thermal unit
        SU (dict): The startup rate of the thermal unit
        TU (dict): The minimum uptime of thermal units

    Returns:
        gp.tupledict: The constraints for the peak up bound

    """
    return model.addConstrs(
        (
            p[unit_g, t] + spin[unit_g, t]
            <= (
                thermal_derated_capacity.loc[t + (step_k - 1) * 24, unit_g]
                - thermal_min_capacity[unit_g]
            )
            * u[unit_g, t]
            - (thermal_derated_capacity.loc[t + (step_k - 1) * 24, unit_g] - SU[unit_g])
            * v[unit_g, t]
            - max(0, (SU[unit_g] - SD[unit_g])) * w[unit_g, t + 1]
            for t in range(1, sim_horizon)  # The constraints index w at t+1
            for unit_g in thermal_units
            if TU[unit_g] == 1
        ),
        name="peakUpBnd",
    )


def add_c_ramp_down_init(
    model: gp.Model,
    p: gp.tupledict,
    w: gp.tupledict,
    thermal_units: list,
    initial_p: dict,
    initial_u: dict,
    thermal_min_capacity: dict,
    RD: dict,
    SD: dict,
) -> gp.tupledict:
    """
    Equation 36 of Kneuven et al (2019) based on Damci-Kurt et al. (2016).
    Ramp-down constraint at t=1. The equation was modified to be expressed
    in terms of power output above the minimum capacity.

    Args:
        model (gp.Model): The optimization model
        p (gp.tupledict): The power output above the minimum capacity
        w (gp.tupledict): The shutdown of the thermal unit
        thermal_units (list): The list of thermal units
        initial_p (dict): The initial power output above the minimum capacity
        initial_u (dict): The initial status of the thermal unit
        thermal_min_capacity (dict): The minimum capacity of the thermal unit
        RD (dict): The ramp-down rate of the thermal unit
        SD (dict): The shutdown rate of the thermal unit

    Returns:
        gp.tupledict: The constraints for the ramp-down at t=1

    """
    # Ramping when t=1 is dependent on the initial condition at t=0
    return model.addConstrs(
        (
            initial_p[unit_g] - p[unit_g, 1]
            <= (SD[unit_g] - thermal_min_capacity[unit_g] - RD[unit_g]) * w[unit_g, 1]
            + RD[unit_g] * initial_u[unit_g]
            for unit_g in thermal_units
        ),
        name="rampDownInit",
    )


def add_c_ramp_down(
    model: gp.Model,
    p: gp.tupledict,
    u: gp.tupledict,
    w: gp.tupledict,
    sim_horizon: int,
    thermal_units: list,
    thermal_min_capacity: dict,
    RD: dict,
    SD: dict,
) -> gp.tupledict:
    """Equation 36 of Kneuven et al (2019) based on Damci-Kurt et al. (2016).
    Ramp-down constraint at t>1. The equation was modified to be expressed
    in terms of power output above the minimum capacity.

    Args:
        model (gp.Model): The optimization model
        p (gp.tupledict): The power output above the minimum capacity
        u (gp.tupledict): The status of the thermal unit
        w (gp.tupledict): The shutdown of the thermal unit
        sim_horizon (int): The simulation horizon
        thermal_units (list): The list of thermal units
        thermal_min_capacity (dict): The minimum capacity of the thermal unit
        RD (dict): The ramp-down rate of the thermal unit
        SD (dict): The shutdown rate of the thermal unit

    Returns:
        gp.tupledict: The constraints for the ramp-down at t>1

    """
    return model.addConstrs(
        (
            p[unit_g, t - 1] - p[unit_g, t]
            <= (SD[unit_g] - thermal_min_capacity[unit_g] - RD[unit_g]) * w[unit_g, t]
            + RD[unit_g] * u[unit_g, t - 1]
            for unit_g in thermal_units
            for t in range(2, sim_horizon + 1)
        ),
        name="rampDown",
    )


def add_c_ramp_up_init(
    model: gp.Model,
    pbar: gp.tupledict,
    u: gp.tupledict,
    v: gp.tupledict,
    thermal_units: list,
    initial_p: dict,
    thermal_min_capacity: dict,
    RU: dict,
    SU: dict,
) -> gp.tupledict:
    """Equation 35 of Kneuven et al (2019) based on Damci-Kurt et al. (2016).
    Ramp-up constraint at t=1. The equation was modified to be expressed
    in terms of power output above the minimum capacity.

    Args:
        model (gp.Model): The optimization model
        pbar (gp.tupledict): The maximum power output above the minimum capacity
        u (gp.tupledict): The status of the thermal unit
        v (gp.tupledict): The startup of the thermal unit
        thermal_units (list): The list of thermal units
        initial_p (dict): The initial power output above the minimum capacity
        thermal_min_capacity (dict): The minimum capacity of the thermal unit
        RU (dict): The ramp-up rate of the thermal unit
        SU (dict): The startup rate of the thermal unit

    Returns:
        gp.tupledict: The constraints for the ramp-up at t=1

    """
    # Ramp up when t=1 is dependent on the initial condition at t=0
    return model.addConstrs(
        (
            pbar[unit_g, 1] - initial_p[unit_g]
            <= (SU[unit_g] - thermal_min_capacity[unit_g] - RU[unit_g]) * v[unit_g, 1]
            + RU[unit_g] * u[unit_g, 1]
            for unit_g in thermal_units
        ),
        name="rampUpInit",
    )


def add_c_ramp_up(
    model: gp.Model,
    p: gp.tupledict,
    pbar: gp.tupledict,
    u: gp.tupledict,
    v: gp.tupledict,
    sim_horizon: int,
    thermal_units: list,
    thermal_min_capacity: dict,
    RU: dict,
    SU: dict,
) -> gp.tupledict:
    """Equation 35 of Kneuven et al (2019) based on Damci-Kurt et al. (2016).
    Ramp-up constraint at t>1. The equation was modified to be expressed
    in terms of power output above the minimum capacity.

    Args:
        model (gp.Model): The optimization model
        p (gp.tupledict): The power output above the minimum capacity
        pbar (gp.tupledict): The maximum power output above the minimum capacity
        u (gp.tupledict): The status of the thermal unit
        v (gp.tupledict): The startup of the thermal unit
        sim_horizon (int): The simulation horizon
        thermal_units (list): The list of thermal units
        thermal_min_capacity (dict): The minimum capacity of the thermal unit
        RU (dict): The ramp-up rate of the thermal unit
        SU (dict): The startup rate of the thermal unit

    """
    return model.addConstrs(
        (
            pbar[unit_g, t] - p[unit_g, t - 1]
            <= (SU[unit_g] - thermal_min_capacity[unit_g] - RU[unit_g]) * v[unit_g, t]
            + RU[unit_g] * u[unit_g, t]
            for unit_g in thermal_units
            for t in range(2, sim_horizon + 1)
        ),
        name="rampUp",
    )


"""Functions for power flow"""


def add_c_link_shortfall(
    model: gp.Model,
    load_shortfall: gp.tupledict,
    p_pmismatch: gp.tupledict,
    n_pmismatch: gp.tupledict,
    timesteps: range,
) -> gp.tupledict:
    """Equation 66 of Kneuven et al (2019). Supply-demand is balanced with an equality constraint.
    To prevent model infeasibility, we introduce a variable to represent the load shortfall s_n(t),
    which can be expressed in terms of non-negative variables: the positive and negative power mismatch.

    Formula: s_n(t) = s_pos(t) - s_neg(t)

    Args:
        model (gp.Model): The optimization model
        load_shortfall (gp.tupledict): The load shortfall
        p_pmismatch (gp.tupledict): The positive power mismatch
        n_pmismatch (gp.tupledict): The negative power mismatch
        timesteps (range): The range of timesteps

    Returns:
        gp.tupledict: The constraints for the load shortfall

    """
    return model.addConstrs(
        load_shortfall[t] == p_pmismatch[t] - n_pmismatch[t] for t in timesteps
    )


def add_c_ref_node(
    model: gp.Model,
    theta: gp.tupledict,
    timesteps: range,
    max_demand_node: str,
) -> gp.tupledict:
    """Equation 64d of Kneuven et al (2019).
    Set the voltage angle at a reference node to zero. The reference node is typically the node with the highest demand.

    Args:
        model (gp.Model): The optimization model
        theta (gp.tupledict): The voltage angle
        timesteps (range): The range of timesteps
        max_demand_node (str): The node with the highest demand

    Returns:
        gp.tupledict: The constraints for the reference node
    """
    return model.addConstrs(
        (theta[max_demand_node, t] == 0 for t in timesteps),
        name="refNode",
    )


def add_c_angle_diff(
    model: gp.Model,
    flow: gp.tupledict,
    theta: gp.tupledict,
    timesteps: range,
    step_k: int,
    edges: list,
    susceptance: pd.DataFrame,
) -> gp.tupledict:
    """Equation 64a of Kneuven et al (2019) expresses the power flow in a transmission line
    as a function of the voltage angle difference between the two buses it connects.
    In DC-OPF, the real power flow in a transmission line is directly proportional
    to the difference in voltage angles.

    Note: If the voltage-angle variable is not implemented, then we can use the Kirchhoff formulation.

    Args:
        model (gp.Model): The optimization model
        flow (gp.tupledict): The power flow
        theta (gp.tupledict): The voltage angle
        timesteps (range): The range of timesteps
        step_k (int): The current iteration
        edges (list): The list of edges
        susceptance (pd.DataFrame): The susceptance matrix

    Returns:
        gp.tupledict: The constraints for the angle difference

    """
    return model.addConstrs(
        (
            flow[a, b, t]
            == susceptance.loc[t + (step_k - 1) * 24, (a, b)]
            * (theta[a, t] - theta[b, t])
            for (a, b) in edges
            for t in timesteps
        ),
        name="angleDiff",
    )


def add_c_kirchhoff(
    model: gp.Model,
    flow: gp.tupledict,
    timesteps: range,
    step_k: int,
    edges: list,
    cycle_map: dict,
    susceptance: pd.DataFrame,
) -> gp.tupledict:
    """Equation 23b in Horsch et al (2018). This constraint implements
    the Kirchhoff circuit laws directly on the flow variables.

    Note: This constraint replaces the voltage-angle difference formulation.

    Args:
        model (gp.Model): The optimization model
        flow (gp.tupledict): The power flow variable
        timesteps (range): The range of timesteps
        step_k (int): The current iteration
        edges (list): The list of edges
        cycle_map (dict): The cycle map (created by DataProcessor class)
        susceptance (pd.DataFrame): The susceptance matrix

    Returns:
        gp.tupledict: The constraints for the Kirchhoff circuit laws

    """
    cycle_incidence = pd.DataFrame(
        0,
        index=pd.MultiIndex.from_tuples(edges, names=["source", "sink"]),
        columns=cycle_map.keys(),
    )

    # The flow is positive with respect to the edges specified by the user
    for cycle_id, cycle in cycle_map.items():
        # Kirchhoff formulation requires summing along the loop direction.
        # If an arc is in cycle_graph and the input file, then 1. Otherwise, -1.
        cycle_graph = nx.path_graph(cycle)
        cycle_edges = list(cycle_graph.edges)
        # Append the last arc connecting the last node to the first node
        cycle_edges.append((cycle_edges[-1][-1], cycle_edges[0][0]))

        # We also need the reactance (susceptance) to calculate the factors in the cycle-incidence
        cycle_susceptance = pd.DataFrame()
        for cycle_edge in cycle_edges:
            if cycle_edge in edges:
                cycle_incidence.loc[cycle_edge, cycle_id] = 1
                cycle_susceptance = pd.concat(
                    [cycle_susceptance, susceptance[cycle_edge]], axis=1
                )
            else:
                cycle_incidence.loc[(cycle_edge[1], cycle_edge[0]), cycle_id] = -1
                cycle_susceptance = pd.concat(
                    [
                        cycle_susceptance,
                        susceptance[(cycle_edge[1], cycle_edge[0])],
                    ],
                    axis=1,
                )

        return model.addConstrs(
            (
                gp.quicksum(
                    (
                        cycle_incidence.loc[(a, b), cycle_id]
                        * 1
                        / cycle_susceptance.loc[t + (step_k - 1) * 24, [(a, b)]].values[
                            0
                        ]
                        * flow[a, b, t]
                    )
                    for (a, b) in cycle_susceptance.columns
                )
                == 0
                for t in timesteps
            ),
            name=f"kirchhoff_{cycle_id}",
        )


def add_c_flow_balance(
    model: gp.Model,
    pthermal: gp.tupledict,
    phydro: gp.tupledict,
    psolar: gp.tupledict,
    pwind: gp.tupledict,
    pimp: gp.tupledict,
    pcharge: gp.tupledict,
    pdis: gp.tupledict,
    pos_pmismatch: gp.tupledict,
    neg_pmismatch: gp.tupledict,
    flow: gp.tupledict,
    timesteps: range,
    step_k: int,
    thermal_units: list,
    hydro_units: list,
    solar_units: list,
    wind_units: list,
    import_units: list,
    nodes: list,
    node_edge: dict,
    node_generator: dict,
    charge_storage: dict,
    discharge_storage: dict,
    demand_nodes: list,
    demand: pd.DataFrame,
    line_loss_factor: float,
) -> gp.tupledict:
    """Equation 65 of Kneuven et al (2019). Supply-demand is balanced with an equality constraint.
    The power flow in the transmission line is also considered.

    Args:
        model (gp.Model): The optimization model
        pthermal (gp.tupledict): The power output of thermal units
        phydro (gp.tupledict): The power output of hydro units
        psolar (gp.tupledict): The power output of solar units
        pwind (gp.tupledict): The power output of wind units
        pimp (gp.tupledict): The power output of import units
        pcharge (gp.tupledict): The charge of energy storage units
        pdis (gp.tupledict): The discharge of energy storage units
        p_pmismatch (gp.tupledict): The positive power mismatch
        n_pmismatch (gp.tupledict): The negative power mismatch
        flow (gp.tupledict): The power flow
        timesteps (range): The range of timesteps
        step_k (int): The current iteration
        thermal_units (list): The list of thermal units
        hydro_units (list): The list of hydro units
        solar_units (list): The list of solar units
        wind_units (list): The list of wind units
        import_units (list): The list of import units
        nodes (list): The list of nodes
        node_edge (dict): The edges connected to a node
        node_generator (dict): The generators connected to a node
        charge_ess (dict): Storage units to charge from this node
        discharge_ess (dict): Storage units to discharge to this node
        demand_nodes (list): The list of demand nodes
        demand (pd.DataFrame): The demand data
        line_loss_factor (float): The transmission line loss factor

    Returns:
        gp.tupledict: The constraints for the power flow balance

    """

    def get_unit_generation(unit: str, t: int):
        if unit in thermal_units:
            return pthermal[unit, t]
        elif unit in hydro_units:
            return phydro[unit, t]
        elif unit in solar_units:
            return psolar[unit, t]
        elif unit in wind_units:
            return pwind[unit, t]
        elif unit in import_units:
            return pimp[unit, t]
        else:
            raise ValueError(
                f"PowNet: Unit {unit} not found in any of the generation types but is connected to the node."
            )

    constraints = gp.tupledict()
    # Impose a line loss factor to account for the inefficiency of the transmission line
    line_efficiency = 1 - line_loss_factor
    for t in timesteps:
        for node in nodes:
            generation = 0
            # Loops through generators located *in* the node (aggregated generation)
            for unit_g in node_generator[node]:
                generation += get_unit_generation(unit_g, t)

            # A grid storage system charges from this node
            storage_charge = 0
            if node in charge_storage:
                storage_systems = charge_storage[node]
                for storage_system in storage_systems:
                    storage_charge += pcharge[storage_system, t]

            # If an ESS is to inject into this node
            storage_discharge = 0
            if node in discharge_storage:
                storage_systems = discharge_storage[node]
                for storage_system in storage_systems:
                    storage_discharge += pdis[storage_system, t]

            # Get the demand of node n at time t
            if node in demand_nodes:
                demand_n_t = demand.loc[t + (step_k - 1) * 24, node]
                mismatch = pos_pmismatch[node, t] - neg_pmismatch[node, t]
            else:
                demand_n_t = 0
                mismatch = 0

            # Flow into a node is positive, while flow out is negative
            arc_flow = 0
            for x, y in node_edge[node]:
                if x == node:
                    arc_flow -= flow[x, y, t]
                elif y == node:
                    arc_flow += flow[x, y, t]

            # Given the above terms, we can specify the energy balance
            cname = f"flowBal[{node},{t}]"
            constraints[cname] = model.addConstr(
                (
                    generation * line_efficiency
                    + arc_flow
                    + mismatch
                    + storage_discharge  # Already factored in discharge efficiency is the ESS balance
                    == demand_n_t + storage_charge
                ),
                name=cname,
            )
    return constraints


"""Functions for the spinning-reserve requirement"""


def add_c_reserve_req_1(
    model: gp.Model,
    spin: gp.tupledict,
    spin_shortfall: gp.tupledict,
    timesteps: range,
    step_k: int,
    thermal_units: list,
    spin_requirement: pd.DataFrame,
) -> gp.tupledict:
    """Equation 68 of Kneuven et al (2019) based on Morales-España et al. (2013).
    System-wide spinning reserve requirement. The spinning reserve is the sum of
    the spinning reserve of all thermal units.

    NOTE: This formulation is suitable when we implement the spinning-reserve variable (spin).
    Only one reserve_requirement is constraint is needed.

    Args:
        model (gp.Model): The optimization model
        spin (gp.tupledict): The spinning reserve
        spin_shortfall (gp.tupledict): The spinning reserve shortfall
        timesteps (range): The range of timesteps
        step_k (int): The current iteration
        thermal_units (list): The list of thermal units
        spin_requirement (pd.DataFrame): The spinning reserve requirement at each hour (MW)

    Returns:
        gp.tupledict: The constraints for the spinning reserve requirement


    """
    return model.addConstrs(
        (
            gp.quicksum(spin[unit_g, t] for unit_g in thermal_units) + spin_shortfall[t]
            >= spin_requirement[t + (step_k - 1) * 24]
            for t in timesteps
        ),
        name="reserveReq1",
    )


def add_c_reserve_req_2(
    model: gp.Model,
    pbar: gp.tupledict,
    u: gp.tupledict,
    spin_shortfall: gp.tupledict,
    timesteps: range,
    step_k: int,
    thermal_units: list,
    thermal_min_capacity: dict,
    demand_nodes: list,
    demand: pd.DataFrame,
    spin_requirement: pd.Series,
) -> gp.tupledict:
    """Equation 67 of Kneuven et al (2019) based on Carrion and Arroyo (2006)
    and Ostrowski et al. (2012). The spinning reserve is expressed in terms of the
    maximum power output of all thermal units. Since the pbar in our implementation is
    actuallly pbar_tilde in Knueven et al (2019), we substitute pbar with pbar + u*min_capacity.

    NOTE: This formulation is suitable when we implement the pbar variable. Only one reserve_requirement
    is constraint is needed.

    Args:
        model (gp.Model): The optimization model
        pbar (gp.tupledict): The maximum power output above the minimum capacity
        u (gp.tupledict): The status of the thermal unit
        spin_shortfall (gp.tupledict): The spinning reserve shortfall
        timesteps (range): The range of timesteps
        step_k (int): The current iteration
        thermal_units (list): The list of thermal units
        thermal_min_capacity (dict): The minimum capacity of the thermal unit
        demand_nodes (list): The list of demand nodes
        demand (pd.DataFrame): The demand data
        spin_requirement (pd.Series): The spinning reserve requirement at each hour (MW)

    Returns:
        gp.tupledict: The constraints for the spinning reserve requirement

    """
    return model.addConstrs(
        (
            gp.quicksum(
                pbar[unit_g, t] + thermal_min_capacity[unit_g] * u[unit_g, t]
                for unit_g in thermal_units
            )
            + spin_shortfall[t]
            >= gp.quicksum(demand.loc[t + (step_k - 1) * 24, n] for n in demand_nodes)
            + spin_requirement[t + (step_k - 1) * 24]
            for t in timesteps
        ),
        name="reserveReq2",
    )


"""Functions for renewables and others"""


def add_c_hydro_limit_daily(
    model: gp.Model,
    phydro: gp.tupledict,
    step_k: int,
    sim_horizon: int,
    hydro_units: list,
    hydro_capacity: pd.DataFrame,
) -> gp.tupledict:
    """
    Defines the daily limit of hydro generation. Assumes that a certain amount of water is
    available for hydropower generation each day. In this case, the dataframe hydro_capacity has a length of 365 days
    instead of 8760 hours.

    Args:
        model (gp.Model): The optimization model
        phydro (gp.tupledict): The power output of hydro units
        step_k (int): The current iteration
        sim_horizon (int): The simulation horizon
        hydro_units (list): The list of hydro units
        hydro_capacity (pd.DataFrame): The capacity of the hydro unit

    Returns:
        gp.tupledict: The constraints for the daily hydro limit

    Raises:
        ValueError: If the simulation horizon is not divisible by 24

    """
    # When formulating with daily hydropower, sim_horizon must be divisible
    # by 24 because the hydro_capacity is daily.

    if sim_horizon % 24 != 0:
        raise ValueError(
            "The simulation horizon must be divisible by 24 when using daily hydropower capacity."
        )
    constraints = gp.tupledict()
    max_day = sim_horizon // 24
    for day in range(step_k, step_k + max_day):
        for hydro_unit in hydro_units:
            cname = f"hydro_limit_daily[{hydro_unit},{day}]"
            current_day = day - step_k + 1
            constraints[cname] = model.addConstr(
                gp.quicksum(
                    phydro[hydro_unit, t]
                    for t in range(1 + (current_day - 1) * 24, current_day * 24 + 1)
                )
                <= hydro_capacity.loc[day, hydro_unit],
                name=cname,
            )
    return constraints


def add_c_link_unit_pu(
    model: gp.Model,
    pdispatch: gp.tupledict,
    u: gp.tupledict,
    type: str,  # "wind" or "solar" or "hydro" or "import"
    timesteps: range,
    step_k: int,
    units: list,
    capacity_df: pd.DataFrame,
) -> gp.tupledict:
    return model.addConstrs(
        (
            pdispatch[unit, t]
            <= u[unit, t] * get_capacity_value(t, unit, step_k, capacity_df)
            for unit in units
            for t in timesteps
        ),
        name=f"link_{type}_pu",
    )


def add_c_link_unit_pu_constant(
    model: gp.Model,
    pdispatch: gp.tupledict,
    u: gp.tupledict,
    timesteps: range,
    units: list,
    contracted_capacity: dict[str, float],
) -> gp.tupledict:
    """Similar to add_c_link_unit_pu, but the upper bound has a constant capacity"""
    return model.addConstrs(
        pdispatch[unit, t] <= u[unit, t] * contracted_capacity[unit]
        for unit in units
        for t in timesteps
    )


def add_c_unit_curtail_ess(
    model: gp.Model,
    pdispatch: gp.tupledict,
    pcurtail: gp.tupledict,
    pcharge: gp.tupledict,
    type: str,  # "wind" or "solar"
    timesteps: range,
    step_k: int,
    units: list,
    capacity_df: pd.DataFrame,
    ess_attached: dict[str, list[str]],
) -> gp.tupledict:

    constraints = gp.tupledict()
    for unit in units:
        has_storage = unit in ess_attached
        for t in timesteps:
            pcharge_unit_t = 0
            if has_storage:
                # A unit may have multiple storage systems
                for storage_unit in ess_attached[unit]:
                    pcharge_unit_t += pcharge[storage_unit, t]

            cname = f"{type}_curtail_ess[{unit},{t}]"
            constraints[cname] = model.addConstr(
                (
                    pdispatch[unit, t] + pcurtail[unit, t] + pcharge_unit_t
                    == get_capacity_value(t, unit, step_k, capacity_df)
                ),
                name=cname,
            )
    return constraints


def add_c_unit_curtail_ess_daily(
    model: gp.Model,
    pdispatch: gp.tupledict,
    pcurtail: gp.tupledict,
    pcharge: gp.tupledict,
    type: str,  # "wind" or "solar"
    sim_horizon: int,
    step_k: int,
    units: list,
    capacity_df: pd.DataFrame,
    ess_attached: dict[str, list[str]],
) -> gp.tupledict:

    constraints = gp.tupledict()
    max_day = sim_horizon // 24
    for unit in units:
        has_storage = unit in ess_attached
        for day in range(step_k, step_k + max_day):

            current_day = day - step_k + 1
            timesteps_in_day = range(1 + (current_day - 1) * 24, current_day * 24 + 1)

            pcharge_unit_day = 0
            if has_storage:
                # A unit may have multiple storage systems
                for storage_unit in ess_attached[unit]:
                    pcharge_unit_day += gp.quicksum(
                        pcharge[storage_unit, t] for t in timesteps_in_day
                    )

            cname = f"{type}_curtail_ess[{unit},{day}]"
            constraints[cname] = model.addConstr(
                (
                    gp.quicksum(
                        pdispatch[unit, t] + pcurtail[unit, t] for t in timesteps_in_day
                    )
                    + pcharge_unit_day
                    == capacity_df.loc[day, unit]
                ),
                name=cname,
            )
    return constraints


""" Energy storage constraints
"""


def add_c_link_ess_charge(
    model: gp.Model,
    pcharge: gp.tupledict,
    ucharge: gp.tupledict,
    timesteps: range,
    units: list,
    max_charge: dict[str, float],
) -> gp.tupledict:
    return model.addConstrs(
        (
            pcharge[unit, t] <= ucharge[unit, t] * max_charge[unit]
            for unit in units
            for t in timesteps
        ),
        name="link_ess_charge",
    )


def add_c_link_ess_discharge(
    model: gp.Model,
    pdischarge: gp.tupledict,
    udischarge: gp.tupledict,
    timesteps: range,
    units: list,
    max_discharge: dict[str, float],
) -> gp.tupledict:
    return model.addConstrs(
        (
            pdischarge[unit, t] <= udischarge[unit, t] * max_discharge[unit]
            for unit in units
            for t in timesteps
        ),
        name="link_ess_discharge",
    )


def add_c_link_ess_state(
    model: gp.Model,
    ucharge: gp.tupledict,
    udischarge: gp.tupledict,
    timesteps: range,
    units: list,
) -> gp.tupledict:
    return model.addConstrs(
        (
            ucharge[unit, t] + udischarge[unit, t] <= 1
            for unit in units
            for t in timesteps
        ),
        name="link_ess_state",
    )


def add_c_unit_ess_balance_init(
    model: gp.Model,
    pcharge: gp.tupledict,
    pdischarge: gp.tupledict,
    charge_state: gp.tupledict,
    units: list,
    charge_state_init: dict[str, float],
    charge_efficiency: dict[str, float],
    discharge_efficiency: dict[str, float],
    self_discharge_rate: dict[str, float],
) -> gp.tupledict:
    t = 1
    return model.addConstrs(
        (
            charge_state[unit, t]
            == (1 - self_discharge_rate[unit]) * charge_state_init[unit]
            + charge_efficiency[unit] * pcharge[unit, t]
            - pdischarge[unit, t] / discharge_efficiency[unit]
            for unit in units
        ),
        name="unit_ess_balance_init",
    )


def add_c_unit_ess_balance(
    model: gp.Model,
    pcharge: gp.tupledict,
    pdischarge: gp.tupledict,
    charge_state: gp.tupledict,
    units: list,
    sim_horizon: int,
    charge_efficiency: dict[str, float],
    discharge_efficiency: dict[str, float],
    self_discharge_rate: dict[str, float],
) -> gp.tupledict:
    return model.addConstrs(
        (
            charge_state[unit, t]
            == (1 - self_discharge_rate[unit]) * charge_state[unit, t - 1]
            + charge_efficiency[unit] * pcharge[unit, t]
            - pdischarge[unit, t] / discharge_efficiency[unit]
            for unit in units
            for t in range(2, sim_horizon + 1)
        ),
        name="unit_ess_balance",
    )



"""visualizer.py: This module contains the Visualizer class, which provides methods to visualize the output from PowNet.
"""

import os

import contextily as cx
import geopandas as gpd
import matplotlib as mpl
from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from pownet.data_utils import get_fuel_color_map, create_geoseries_columns


class Visualizer:
    def __init__(self, model_id: str) -> None:
        self.model_id: str = model_id
        self.fuel_color_map: dict = get_fuel_color_map()

    def plot_fuelmix_bar(
        self,
        dispatch: pd.DataFrame,
        demand: pd.Series,
        output_folder: str = None,
    ) -> None:
        """Create a bar plot of the fuel mix.

        Args:
            dispatch (pd.DataFrame): The dispatch of each generator.
            demand (pd.Series): The demand of the system.
            output_folder (str): If specified, then the plot is saved in the folder.

        Returns:
            None
        """
        # Use total_timesteps to index demand because
        # the length of demand can be longer than the simulation hours
        total_timesteps: int = dispatch.shape[0]

        # Plotting section
        fig, ax = plt.subplots(figsize=(8, 5))

        dispatch.plot.bar(
            stacked=True, ax=ax, linewidth=0, color=self.fuel_color_map, legend=False
        )
        ax.plot(
            range(0, total_timesteps),
            demand[:total_timesteps],
            color="k",
            linewidth=2,
            linestyle=":",
            label="demand",
        )
        ax.set_xlabel(dispatch.index.name)

        # Plot formatting
        legend = fig.legend(
            loc="outside lower center",
            ncols=5,
            fontsize="small",
            bbox_to_anchor=(0.5, -0.12),
        )
        ax.set_ylabel("Power (MW)")
        ax.set_ylim(top=(demand[:total_timesteps].max() * 1.30).values[0])

        if output_folder is not None:
            figure_name = f"{self.model_id}_fuelmix.png"
            fig.savefig(
                os.path.join(output_folder, figure_name),
                bbox_extra_artists=(legend,),
                bbox_inches="tight",
                dpi=350,
            )
        plt.show()

    def plot_fuelmix_area(
        self,
        dispatch: pd.DataFrame,
        demand: pd.Series,
        output_folder: str = None,
    ) -> None:
        """Create an area plot of the fuel mix.

        Args:
            dispatch (pd.DataFrame): The dispatch of each generator.
            demand (pd.Series): The demand of the system.
            output_folder (str): If specified, then the plot is saved in the folder.

        Returns:
            None

        """
        # Use total_timesteps to index demand because
        # the length of demand can be longer than the total simulation hours
        total_timesteps: int = dispatch.shape[0]

        fig, ax = plt.subplots(figsize=(8, 5))
        dispatch.plot.area(
            stacked=True, ax=ax, linewidth=0, color=self.fuel_color_map, legend=False
        )
        ax.plot(
            demand[:total_timesteps],
            color="k",
            linewidth=2,
            linestyle=":",
            label="demand",
        )
        ax.set_xlabel(dispatch.index.name)

        legend = fig.legend(
            loc="outside lower center",
            ncols=5,
            fontsize="small",
            bbox_to_anchor=(0.5, -0.12),
        )
        ax.set_ylabel("Power (MW)")
        ax.set_ylim(top=(demand[:total_timesteps].max() * 1.30).values[0])

        if output_folder is not None:
            figure_name = f"{self.model_id}_fuelmix.png"
            fig.savefig(
                os.path.join(output_folder, figure_name),
                bbox_extra_artists=(legend,),
                bbox_inches="tight",
                dpi=350,
            )
        plt.show()

    def plot_thermal_units(
        self,
        thermal_dispatch: pd.DataFrame,
        unit_status: pd.DataFrame,
        thermal_rated_capacity: dict[str, float],
        output_folder: str = None,
    ) -> None:
        """Plot the on/off status of individual thermal units

        Args:
            thermal_dispatch (pd.DataFrame): The dispatch of each thermal unit.
            unit_status (pd.DataFrame): The status of each thermal unit.
            thermal_rated_capacity (dict[str, float]): Rated capacity of each thermal unit.
            output_folder (str): If specified, then the plot is saved in the folder.

        Returns:
            None

        """
        thermal_units = thermal_dispatch["node"].unique()

        for unit in thermal_units:
            # Extract the dispatch of each thermal unit and plot the value
            df1 = thermal_dispatch[thermal_dispatch.node == unit]
            df2 = unit_status[unit_status["node"] == unit]

            fig, ax1 = plt.subplots(figsize=(8, 5))
            ax2 = ax1.twinx()

            ax1.step(df1["hour"], df1["value"], where="mid", color="b", label="Power")
            # If ymax is too low, then we cannot see the blue line
            ax1.set_ylim(bottom=0, top=thermal_rated_capacity[unit] * 1.05)
            ax1.tick_params(axis="x", labelrotation=45)
            ax1.set_xlabel("Hour")
            ax1.set_ylabel("Power (MW)")

            ax2.bar(
                df2["hour"], df2["value"], color="k", alpha=0.2, label="Unit status"
            )
            ax2.set_ylim(bottom=0, top=1)
            ax2.set_ylabel("Unit Status")
            plt.title(unit)

            if output_folder is not None:
                unit_plot_folder = os.path.join(
                    output_folder, f"{self.model_id}_unit_plots"
                )
                if not os.path.exists(unit_plot_folder):
                    os.mkdir(unit_plot_folder)

                fig.savefig(
                    os.path.join(unit_plot_folder, f"{unit}.png"),
                    dpi=350,
                )
            plt.show()

    def plot_mean_thermal_unit_hourly_status(
        self, thermal_unit_mean_hourly_status: pd.DataFrame, output_folder: str = None
    ) -> None:
        fig, ax = plt.subplots(figsize=(8, 5))
        ax.imshow(thermal_unit_mean_hourly_status.T, cmap="binary", aspect="auto")
        ax.set_xlabel("Hour of the day")
        ax.set_ylabel("Thermal unit ID")
        ax.set_title("Mean hourly status of thermal units")

        # Create a colorbar
        norm = mpl.colors.Normalize(vmin=0, vmax=1)
        cbar = plt.colorbar(mpl.cm.ScalarMappable(norm=norm, cmap="binary"), ax=ax)
        cbar.set_label("Unit status")

        legend = fig.legend(
            loc="outside lower center",
            fontsize="small",
            bbox_to_anchor=(0.5, -0.12),
        )

        if output_folder is not None:
            figure_name = f"{self.model_id}_thermal_unit_status.png"
            fig.savefig(
                os.path.join(output_folder, figure_name),
                bbox_extra_artists=(legend,),
                bbox_inches="tight",
                dpi=350,
            )
        plt.show()

    def plot_lmp(
        self,
        lmp_df: pd.DataFrame,
        output_folder: str = None,
        max_ylim: float = 200,
    ) -> None:
        """Plots unique locational marginal price (LMP) timeseries.
        For each unique LMP timeseries, a representative node is chosen
        based on ordering in the dataframe.

        Args:
            lmp_df (pd.DataFrame): LMP timeseries.
            output_folder (str): If specified, then the plot is saved in the folder.
            max_ylim (float): Maximum y-axis limit.

        Returns:
            None

        """
        # Find uni
        unique_lmp = lmp_df.copy().T.drop_duplicates().T

        fig, ax = plt.subplots()
        unique_lmp.plot(ax=ax, linewidth=2)
        ax.set_xlabel("Hour")
        ax.set_ylabel("LMP ($/MWh)")
        # Place legend at the bottom
        ax.legend(loc="upper center", bbox_to_anchor=(0.5, -0.1), ncol=4)
        ax.set_ylim(top=max_ylim)

        # Adjust layout to prevent legend from being cut off
        plt.tight_layout()
        plt.subplots_adjust(bottom=0.2)

        if output_folder is not None:
            figure_name = f"{self.model_id}_lmp.png"
            fig.savefig(
                os.path.join(output_folder, figure_name),
                dpi=350,
            )
        plt.show()

    def plot_line_usage(
        self,
        max_line_usage: pd.DataFrame,
        output_folder: str,
    ) -> None:
        """Flow variables must have the max_line_usage column"""
        max_line_usage = create_geoseries_columns(max_line_usage)
        # There should be three geoseries columns: source_location,
        # sink_location, and geometry.
        max_line_usage_gdf = gpd.GeoDataFrame(
            max_line_usage, geometry="geometry", crs="EPSG:4326"
        )

        def get_linewidth(capacity):
            min_capacity = max_line_usage.rated_capacity.min()
            max_capacity = max_line_usage.rated_capacity.max()
            min_linewidth = 2
            max_linewidth = 6
            # Scale capacity to between 1 and 10 to avoid log(0) errors
            scaled_capacity = 1 + 9 * (capacity - min_capacity) / (
                max_capacity - min_capacity
            )
            log_capacity = np.log10(scaled_capacity)
            # Scale the log value to the desired linewidth range.
            return min_linewidth + (log_capacity / np.log10(10)) * (
                max_linewidth - min_linewidth
            )

        max_line_usage_gdf["linewidth"] = max_line_usage_gdf["rated_capacity"].apply(
            get_linewidth
        )

        # Bin the 'line_usage' column into 5 quantiles for plotting
        # colors = ["#000004", "#3b0f6f", "#8c2981", "#dd4968", "#fd9347"]
        colors = ["#fdd0a2", "#fdae6b", "#fd8d3c", "#e6550d", "#a63603"]
        labels = ["0-20%", "20-40%", "40-60%", "60-80%", "80-100%"]

        bins = [-0.01, 0.2, 0.4, 0.6, 0.8, 1.0]
        max_line_usage_gdf["usage_bin"] = pd.cut(
            max_line_usage_gdf["max_line_usage"],
            bins=bins,
            right=True,
            labels=labels,
        )

        # Plot
        ax = max_line_usage_gdf.plot(
            column="usage_bin",
            figsize=(13, 7),
            cmap=ListedColormap(colors),
            linewidth=max_line_usage_gdf["linewidth"],
            legend=True,
            legend_kwds={
                "title": "Max line capacity rate",
                "loc": "upper left",
                "bbox_to_anchor": (1, 1),
            },
        )
        max_line_usage_gdf["sink_location"].plot(ax=ax, color="k", markersize=5)
        max_line_usage_gdf["source_location"].plot(ax=ax, color="k", markersize=5)

        cx.add_basemap(
            ax, crs=max_line_usage_gdf.crs, source=cx.providers.CartoDB.Positron
        )
        ax.set_axis_off()

        # Format plots
        # Adjust layout to prevent legend from being cut off
        plt.tight_layout()
        plt.subplots_adjust(bottom=0.2)

        if output_folder is not None:
            figure_name = f"{self.model_id}_line_usage.png"
            fig = ax.get_figure()
            fig.savefig(
                os.path.join(output_folder, figure_name),
                dpi=350,
            )
        plt.show()

    def plot_unit_storage_state(
        self,
        hourly_storage_charge: pd.Series,
        hourly_storage_discharge: pd.Series,
        hourly_storage_state: pd.Series,
        output_folder: str = None,
    ) -> None:
        """
        Plot the hourly activity of the energy storage units.

        Args:
            hourly_storage_charge (pd.Series): Hourly charge data.
            hourly_storage_discharge (pd.Series): Hourly discharge data.
            hourly_storage_state (pd.Series): Hourly storage state data.
        """
        storage_units = hourly_storage_charge.columns
        for unit in storage_units[:2]:
            fig, axes = plt.subplots(
                2, 1, figsize=(10, 6), sharex=True, height_ratios=[3, 1]
            )  # Set height ratios

            # Plot storage fraction
            axes[0].plot(hourly_storage_state.index, hourly_storage_state[unit].values)
            axes[0].set_ylabel("Storage Fraction")

            axes[0].set_title(f"{unit} hourly storage state")

            # Plot charge and discharge as bar chart
            axes[1].bar(
                hourly_storage_charge.index,
                hourly_storage_charge[unit].values,
                label="Charge",
            )
            axes[1].bar(
                hourly_storage_discharge.index,
                -1 * hourly_storage_discharge[unit].values,
                label="Discharge",
            )
            axes[1].set_ylabel("Power (MW)")
            axes[1].set_xlabel("Hour")
            # axes[1].legend()

            plt.tight_layout()

            if output_folder is not None:
                unit_plot_folder = os.path.join(
                    output_folder, f"{self.model_id}_energy_storage_plots"
                )
                if not os.path.exists(unit_plot_folder):
                    os.mkdir(unit_plot_folder)

                fig.savefig(
                    os.path.join(unit_plot_folder, f"{unit}.png"),
                    dpi=350,
                )
            plt.show()

    def plot_generation_by_contracts(self, contract_generation: pd.DataFrame) -> None:
        fig, ax = plt.subplots(figsize=(5, 8))
        df_to_plot = contract_generation.sort_values(by=["value"], ascending=False)
        df_to_plot.plot(ax=ax, kind="bar", linewidth=2, legend=False)
        ax.set_ylabel("Generation (MWh)")
        plt.show()



import os


def get_pownet_dir() -> str:
    """Does not assume the user saves the folder under their home directory"""
    return os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


def get_output_dir() -> str:
    return os.path.join(get_pownet_dir(), "outputs")


def get_input_dir() -> str:
    return os.path.join(get_pownet_dir(), "user_inputs")


def get_temp_dir() -> str:
    return os.path.join(get_pownet_dir(), "temp")


def get_model_dir() -> str:
    return os.path.join(get_pownet_dir(), "model_library")


def get_reservoir_dir(model_name: str) -> str:
    return os.path.join(get_model_dir(), model_name, "reservoir")


def get_reservoir_file(model_name: str, filename: str) -> str:
    return os.path.join(get_reservoir_dir(model_name), f"{filename}")


def get_home_dir() -> str:
    return os.path.expanduser("~")


def get_database_dir() -> str:
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "database")


def get_test_dir() -> str:
    return os.path.join(get_pownet_dir(), "src", "test_pownet")


def delete_all_gurobi_solutions() -> None:
    """Remove all Gurobi solution files from the output folder.
    Use this function at the beginning of the simulation when warmstart is on.
    """
    solution_files = os.listdir(get_output_dir())
    for s_file in solution_files:
        file_extension = os.path.splitext(s_file)[1]
        if ".sol" in file_extension:
            os.remove(os.path.join(get_output_dir(), s_file))


def count_mps_files(instance_folder: str) -> int:
    """Count the number of files ending with .mps"""
    num_instances = 0
    for file in os.listdir(instance_folder):
        if file.endswith(".mps"):
            num_instances += 1
    return num_instances
